(*
(C) Copyright Andreas Viktor Hess, DTU, 2018

All Rights Reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

- Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products
  derived from this software without specific prior written
  permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*)

(*  Title:      StatefulTest.thy
    Author:     Andreas Viktor Hess, DTU
*)

theory StatefulTyping
imports ProtocolTransitionSystem
begin

locale stateful_typed_model = typed_model arity public Ana \<Gamma>
  for arity::"'fun::countable \<Rightarrow> nat"
    and public::"'fun \<Rightarrow> bool"
    and Ana::"('fun,'var::countable) term \<Rightarrow> (('fun,'var) terms \<times> ('fun,'var) terms)"
    and \<Gamma>::"('fun,'var) term \<Rightarrow> ('fun,'atom::finite) term_type"
  +
  fixes pair::"'fun"
  assumes pair_arity: "arity pair = 2"
  and Ana_subst': "\<And>f T \<delta> K M. Ana (Fun f T) = (K,M) \<Longrightarrow> Ana (Fun f T \<cdot> \<delta>) = (K \<cdot>set \<delta>,M \<cdot>set \<delta>)"
begin

section {* Stateful constraints *}

lemma Ana_invar_subst'[simp]: "Ana_invar_subst \<S>"
using Ana_subst' unfolding Ana_invar_subst by force

datatype ('a,'b) statefulstrand_step = 
  Send (trm\<^sub>s\<^sub>s\<^sub>t\<^sub>p: "('a,'b) term")
| Receive (trm\<^sub>s\<^sub>s\<^sub>t\<^sub>p: "('a,'b) term")
| Equality (trm\<^sub>s\<^sub>s\<^sub>t\<^sub>p: "('a,'b) term") (trm_r\<^sub>s\<^sub>s\<^sub>t\<^sub>p: "('a,'b) term")
| Inequality (bvars\<^sub>s\<^sub>s\<^sub>t\<^sub>p: "'b list") (trm\<^sub>s\<^sub>s\<^sub>t\<^sub>p: "('a,'b) term") (trm_r\<^sub>s\<^sub>s\<^sub>t\<^sub>p: "('a,'b) term")
| Insert (trm\<^sub>s\<^sub>s\<^sub>t\<^sub>p: "('a,'b) term") (trm_r\<^sub>s\<^sub>s\<^sub>t\<^sub>p: "('a,'b) term")
| Delete (trm\<^sub>s\<^sub>s\<^sub>t\<^sub>p: "('a,'b) term") (trm_r\<^sub>s\<^sub>s\<^sub>t\<^sub>p: "('a,'b) term")
| InSet (trm\<^sub>s\<^sub>s\<^sub>t\<^sub>p: "('a,'b) term") (trm_r\<^sub>s\<^sub>s\<^sub>t\<^sub>p: "('a,'b) term")
| NotInSet (bvars\<^sub>s\<^sub>s\<^sub>t\<^sub>p: "'b list") (trm\<^sub>s\<^sub>s\<^sub>t\<^sub>p: "('a,'b) term") (trm_r\<^sub>s\<^sub>s\<^sub>t\<^sub>p: "('a,'b) term")
where
  "bvars\<^sub>s\<^sub>s\<^sub>t\<^sub>p (Send _) = []"
| "bvars\<^sub>s\<^sub>s\<^sub>t\<^sub>p (Receive _) = []"
| "bvars\<^sub>s\<^sub>s\<^sub>t\<^sub>p (Equality _ _) = []"
| "bvars\<^sub>s\<^sub>s\<^sub>t\<^sub>p (Insert _ _) = []"
| "bvars\<^sub>s\<^sub>s\<^sub>t\<^sub>p (Delete _ _) = []"
| "bvars\<^sub>s\<^sub>s\<^sub>t\<^sub>p (InSet _ _) = []"
| "trm_r\<^sub>s\<^sub>s\<^sub>t\<^sub>p (Send t) = t"
| "trm_r\<^sub>s\<^sub>s\<^sub>t\<^sub>p (Receive t) = t"

type_synonym ('a,'b) statefulstrand = "('a,'b) statefulstrand_step list"

type_synonym ('a,'b) dbstatelist = "(('a,'b) term \<times> ('a,'b) term) list"

type_synonym ('a,'b) dbstate = "(('a,'b) term \<times> ('a,'b) term) set"

definition ik\<^sub>s\<^sub>s\<^sub>t where "ik\<^sub>s\<^sub>s\<^sub>t A \<equiv> {t. Receive t \<in> set A}"

definition bvars\<^sub>s\<^sub>s\<^sub>t::"('a,'b) statefulstrand \<Rightarrow> 'b set" where
  "bvars\<^sub>s\<^sub>s\<^sub>t S \<equiv> \<Union>(set (map (set \<circ> bvars\<^sub>s\<^sub>s\<^sub>t\<^sub>p) S))"

declare bvars\<^sub>s\<^sub>s\<^sub>t_def[simp]

fun FV\<^sub>s\<^sub>s\<^sub>t\<^sub>p::"('a,'b) statefulstrand_step \<Rightarrow> 'b set" where
  "FV\<^sub>s\<^sub>s\<^sub>t\<^sub>p (Send t) = FV t"
| "FV\<^sub>s\<^sub>s\<^sub>t\<^sub>p (Receive t) = FV t"
| "FV\<^sub>s\<^sub>s\<^sub>t\<^sub>p (Equality t t') = FV t \<union> FV t'"
| "FV\<^sub>s\<^sub>s\<^sub>t\<^sub>p (Inequality X t t') = FV t \<union> FV t' - set X"
| "FV\<^sub>s\<^sub>s\<^sub>t\<^sub>p (Insert t t') = FV t \<union> FV t'"
| "FV\<^sub>s\<^sub>s\<^sub>t\<^sub>p (Delete t t') = FV t \<union> FV t'"
| "FV\<^sub>s\<^sub>s\<^sub>t\<^sub>p (InSet t t') = FV t \<union> FV t'"
| "FV\<^sub>s\<^sub>s\<^sub>t\<^sub>p (NotInSet X t t') = FV t \<union> FV t' - set X"

definition FV\<^sub>s\<^sub>s\<^sub>t::"('a,'b) statefulstrand \<Rightarrow> 'b set" where
  "FV\<^sub>s\<^sub>s\<^sub>t S \<equiv> \<Union>(set (map FV\<^sub>s\<^sub>s\<^sub>t\<^sub>p S))"

declare FV\<^sub>s\<^sub>s\<^sub>t_def[simp]

abbreviation subst_apply_pair (infixl "\<cdot>p" 60) where
  "d \<cdot>p \<theta> \<equiv> (case d of (t,t') \<Rightarrow> (t \<cdot> \<theta>, t' \<cdot> \<theta>))"

abbreviation subst_apply_pairs (infixl "\<cdot>pairs" 60) where
  "M \<cdot>pairs \<theta> \<equiv> (\<lambda>d. d \<cdot>p \<theta>) ` M"

abbreviation to_pair where
  "to_pair d \<equiv> case d of (t,t') \<Rightarrow> Fun pair [t,t']"

fun tr::"('fun,'var) statefulstrand \<Rightarrow> ('fun,'var) dbstatelist \<Rightarrow> ('fun,'var) strands" where
  "tr [] D = {[]}"
| "tr (Send t#S) D = (\<lambda>S. snd\<langle>t\<rangle>#S) ` tr S D"
| "tr (Receive t#S) D = (\<lambda>S. rcv\<langle>t\<rangle>#S) ` tr S D"
| "tr (Equality t t'#S) D = (\<lambda>S. t \<doteq> t'#S) ` tr S D"
| "tr (Inequality X t t'#S) D = (\<lambda>S. \<nexists>X\<langle>t \<doteq> t'\<rangle>#S) ` tr S D"
| "tr (Insert t s#S) D = tr S (List.insert (t,s) D)"
| "tr (Delete t s#S) D = (
    \<Union>Di \<in> set (sublists D). (\<Union>S' \<in> tr S (filter (\<lambda>d. d \<notin> set Di) D).
      {(map (\<lambda>d. to_pair (t,s) \<doteq> to_pair d) Di)@
       (map (\<lambda>d. \<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>) (filter (\<lambda>d. d \<notin> set Di) D))@
       S'}
    ))"
| "tr (InSet t s#S) D = (\<Union>S' \<in> tr S D. (\<Union>d \<in> set D. {to_pair (t,s) \<doteq> to_pair d#S'}))"
| "tr (NotInSet X t s#S) D = (\<lambda>S. (map (\<lambda>d. \<nexists>X\<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>) D)@S) ` tr S D"

fun strand_sem_stateful::
  "('fun,'var) terms \<Rightarrow> ('fun,'var) dbstate \<Rightarrow> ('fun,'var) statefulstrand \<Rightarrow> ('fun,'var) constrsem"
  ("\<lbrakk>_; _; _\<rbrakk>\<^sub>s")
where
  "\<lbrakk>M; D; []\<rbrakk>\<^sub>s = (\<lambda>\<I>. True)"
| "\<lbrakk>M; D; Send t#S\<rbrakk>\<^sub>s = (\<lambda>\<I>. M \<turnstile> t \<cdot> \<I> \<and> \<lbrakk>M; D; S\<rbrakk>\<^sub>s \<I>)"
| "\<lbrakk>M; D; Receive t#S\<rbrakk>\<^sub>s = (\<lambda>\<I>. \<lbrakk>insert (t \<cdot> \<I>) M; D; S\<rbrakk>\<^sub>s \<I>)"
| "\<lbrakk>M; D; Equality t t'#S\<rbrakk>\<^sub>s = (\<lambda>\<I>. t \<cdot> \<I> = t' \<cdot> \<I> \<and> \<lbrakk>M; D; S\<rbrakk>\<^sub>s \<I>)"
| "\<lbrakk>M; D; Inequality X t t'#S\<rbrakk>\<^sub>s =
        (\<lambda>\<I>. (\<forall>\<delta>. dom\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta> = set X \<and> ground (img\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta>) \<longrightarrow> t \<cdot> \<delta> \<cdot> \<I> \<noteq> t' \<cdot> \<delta> \<cdot> \<I>) \<and> \<lbrakk>M; D; S\<rbrakk>\<^sub>s \<I>)"
| "\<lbrakk>M; D; Insert t s#S\<rbrakk>\<^sub>s = (\<lambda>\<I>. \<lbrakk>M; insert ((t,s) \<cdot>p \<I>) D; S\<rbrakk>\<^sub>s \<I>)"
| "\<lbrakk>M; D; Delete t s#S\<rbrakk>\<^sub>s = (\<lambda>\<I>. \<lbrakk>M; D - {((t,s) \<cdot>p \<I>)}; S\<rbrakk>\<^sub>s \<I>)"
| "\<lbrakk>M; D; InSet t s#S\<rbrakk>\<^sub>s = (\<lambda>\<I>. (t,s) \<cdot>p \<I> \<in> D \<and> \<lbrakk>M; D; S\<rbrakk>\<^sub>s \<I>)"
| "\<lbrakk>M; D; NotInSet X t s#S\<rbrakk>\<^sub>s =
        (\<lambda>\<I>. (\<forall>\<delta>. dom\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta> = set X \<and> ground (img\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta>) \<longrightarrow> (t,s) \<cdot>p \<delta> \<cdot>p \<I> \<notin> D) \<and> \<lbrakk>M; D; S\<rbrakk>\<^sub>s \<I>)"

lemmas strand_sem_stateful_induct =
  strand_sem_stateful.induct[case_names Nil ConsSnd ConsRcv ConsEq ConsIneq
                                        ConsIns ConsDel ConsIn ConsNotin]

abbreviation constr_sem_stateful ("_ \<Turnstile>\<^sub>s \<langle>_\<rangle>" 90) where "\<I> \<Turnstile>\<^sub>s \<langle>A\<rangle> \<equiv> \<lbrakk>{}; {}; A\<rbrakk>\<^sub>s \<I>"

fun wf\<^sub>s\<^sub>s\<^sub>t::"'b set \<Rightarrow> ('a,'b) statefulstrand \<Rightarrow> bool" where
  "wf\<^sub>s\<^sub>s\<^sub>t V [] = True"
| "wf\<^sub>s\<^sub>s\<^sub>t V (Receive t#S) = (FV t \<subseteq> V \<and> wf\<^sub>s\<^sub>s\<^sub>t V S)"
| "wf\<^sub>s\<^sub>s\<^sub>t V (Send t#S) = wf\<^sub>s\<^sub>s\<^sub>t (V \<union> FV t) S"
| "wf\<^sub>s\<^sub>s\<^sub>t V (Equality t t'#S) = (FV t' \<subseteq> V \<and> wf\<^sub>s\<^sub>s\<^sub>t (V \<union> FV t) S)"
| "wf\<^sub>s\<^sub>s\<^sub>t V (Inequality _ _ _#S) = wf\<^sub>s\<^sub>s\<^sub>t V S"
| "wf\<^sub>s\<^sub>s\<^sub>t V (Insert t s#S) = (FV t \<subseteq> V \<and> FV s \<subseteq> V \<and> wf\<^sub>s\<^sub>s\<^sub>t V S)"
| "wf\<^sub>s\<^sub>s\<^sub>t V (Delete t s#S) = (FV t \<subseteq> V \<and> FV s \<subseteq> V \<and> wf\<^sub>s\<^sub>s\<^sub>t V S)"
| "wf\<^sub>s\<^sub>s\<^sub>t V (InSet t s#S) = wf\<^sub>s\<^sub>s\<^sub>t (V \<union> FV t \<union> FV s) S"
| "wf\<^sub>s\<^sub>s\<^sub>t V (NotInSet _ _ _#S) = wf\<^sub>s\<^sub>s\<^sub>t V S"

fun setops\<^sub>s\<^sub>s\<^sub>t where
  "setops\<^sub>s\<^sub>s\<^sub>t [] = {}"
| "setops\<^sub>s\<^sub>s\<^sub>t (Insert t s#S) = insert (to_pair (t,s)) (setops\<^sub>s\<^sub>s\<^sub>t S)"
| "setops\<^sub>s\<^sub>s\<^sub>t (Delete t s#S) = insert (to_pair (t,s)) (setops\<^sub>s\<^sub>s\<^sub>t S)"
| "setops\<^sub>s\<^sub>s\<^sub>t (InSet t s#S) = insert (to_pair (t,s)) (setops\<^sub>s\<^sub>s\<^sub>t S)"
| "setops\<^sub>s\<^sub>s\<^sub>t (NotInSet _ t s#S) = insert (to_pair (t,s)) (setops\<^sub>s\<^sub>s\<^sub>t S)"
| "setops\<^sub>s\<^sub>s\<^sub>t (_#S) = setops\<^sub>s\<^sub>s\<^sub>t S"

lemma setops\<^sub>s\<^sub>s\<^sub>t_member_iff:
  "Fun pair [t,s] \<in> setops\<^sub>s\<^sub>s\<^sub>t A \<longleftrightarrow>
    (Insert t s \<in> set A \<or> Delete t s \<in> set A \<or> InSet t s \<in> set A \<or> (\<exists>X. NotInSet X t s \<in> set A))"
  (is "?P \<longleftrightarrow> ?Q")
proof
  show "?P \<Longrightarrow> ?Q" by (induct A rule: setops\<^sub>s\<^sub>s\<^sub>t.induct) auto
  show "?Q \<Longrightarrow> ?P"
  proof (induction A)
    case (Cons a A) thus ?case by (cases a) auto
  qed simp
qed

definition trms\<^sub>s\<^sub>s\<^sub>t where "trms\<^sub>s\<^sub>s\<^sub>t S \<equiv> trm\<^sub>s\<^sub>s\<^sub>t\<^sub>p ` set S \<union> trm_r\<^sub>s\<^sub>s\<^sub>t\<^sub>p ` set S"
declare trms\<^sub>s\<^sub>s\<^sub>t_def[simp]

definition eqs_tfr\<^sub>s\<^sub>s\<^sub>t where
  "eqs_tfr\<^sub>s\<^sub>s\<^sub>t S \<equiv>
    (\<forall>t t'. Equality t t' \<in> set S \<and> (\<exists>\<delta>. Unifier \<delta> t t') \<longrightarrow> \<Gamma> t = \<Gamma> t') \<and>
    (\<forall>X t t'. Inequality X t t' \<in> set S \<longrightarrow> (\<forall>x \<in> (FV t \<union> FV t')-set X. \<exists>a. \<Gamma> (Var x) = TAtom a)) \<and>
    (\<forall>t t'.   Insert t t' \<in> set S \<longrightarrow> (\<forall>x \<in> FV t \<union> FV t'. \<exists>a. \<Gamma> (Var x) = TAtom a)) \<and>
    (\<forall>t t'.   Delete t t' \<in> set S \<longrightarrow> (\<forall>x \<in> FV t \<union> FV t'. \<exists>a. \<Gamma> (Var x) = TAtom a)) \<and>
    (\<forall>X t t'. NotInSet X t t' \<in> set S \<longrightarrow> (\<forall>x \<in> (FV t \<union> FV t')-set X. \<exists>a. \<Gamma> (Var x) = TAtom a))"

definition tfr\<^sub>s\<^sub>s\<^sub>t where "tfr\<^sub>s\<^sub>s\<^sub>t S \<equiv> tfr\<^sub>s\<^sub>t (trms\<^sub>s\<^sub>s\<^sub>t S \<union> setops\<^sub>s\<^sub>s\<^sub>t S) \<and> eqs_tfr\<^sub>s\<^sub>s\<^sub>t S"

fun dbupd\<^sub>s\<^sub>s\<^sub>t::"('f,'v) statefulstrand \<Rightarrow> ('f,'v) subst \<Rightarrow> ('f,'v) dbstate \<Rightarrow> ('f,'v) dbstate"
where
  "dbupd\<^sub>s\<^sub>s\<^sub>t [] I D = D"
| "dbupd\<^sub>s\<^sub>s\<^sub>t (Insert t s#A) I D = dbupd\<^sub>s\<^sub>s\<^sub>t A I (insert ((t,s) \<cdot>p I) D)"
| "dbupd\<^sub>s\<^sub>s\<^sub>t (Delete t s#A) I D = dbupd\<^sub>s\<^sub>s\<^sub>t A I (D - {((t,s) \<cdot>p I)})"
| "dbupd\<^sub>s\<^sub>s\<^sub>t (_#A) I D = dbupd\<^sub>s\<^sub>s\<^sub>t A I D"

lemma strand_sem_ik_mono_stateful:
  "\<lbrakk>M; D; A\<rbrakk>\<^sub>s \<I> \<Longrightarrow> \<lbrakk>M \<union> M'; D; A\<rbrakk>\<^sub>s \<I>"
using ideduct_mono by (induct A arbitrary: M M' D rule: strand_sem_stateful.induct) force+

lemma strand_sem_append_stateful:
  "\<lbrakk>M; D; A@B\<rbrakk>\<^sub>s \<I> \<longleftrightarrow> \<lbrakk>M; D; A\<rbrakk>\<^sub>s \<I> \<and> \<lbrakk>M \<union> (ik\<^sub>s\<^sub>s\<^sub>t A \<cdot>set \<I>); dbupd\<^sub>s\<^sub>s\<^sub>t A \<I> D; B\<rbrakk>\<^sub>s \<I>"
  (is "?P \<longleftrightarrow> ?Q \<and> ?R")
proof -
  have 1: "?P \<Longrightarrow> ?Q" by (induct A rule: strand_sem_stateful.induct) auto

  have 2: "?P \<Longrightarrow> ?R"
  proof (induction A arbitrary: M D B)
    case (Cons a A) thus ?case
    proof (cases a)
      case (Receive t)
      have "insert (t \<cdot> \<I>) (M \<union> (ik\<^sub>s\<^sub>s\<^sub>t A \<cdot>set \<I>)) = M \<union> (ik\<^sub>s\<^sub>s\<^sub>t (a#A) \<cdot>set \<I>)"
           "dbupd\<^sub>s\<^sub>s\<^sub>t A \<I> D = dbupd\<^sub>s\<^sub>s\<^sub>t (a#A) \<I> D"
        using Receive unfolding ik\<^sub>s\<^sub>s\<^sub>t_def by auto
      thus ?thesis using Cons Receive by force
    qed (auto simp add: ik\<^sub>s\<^sub>s\<^sub>t_def)
  qed (simp_all add: strand_sem_ik_mono_stateful)

  have 3: "?Q \<Longrightarrow> ?R \<Longrightarrow> ?P"
  proof (induction A arbitrary: M D)
    case (Cons a A) thus ?case
    proof (cases a)
      case (Receive t)
      have "insert (t \<cdot> \<I>) (M \<union> (ik\<^sub>s\<^sub>s\<^sub>t A \<cdot>set \<I>)) = M \<union> (ik\<^sub>s\<^sub>s\<^sub>t (a#A) \<cdot>set \<I>)"
           "dbupd\<^sub>s\<^sub>s\<^sub>t A \<I> D = dbupd\<^sub>s\<^sub>s\<^sub>t (a#A) \<I> D"
        using Receive unfolding ik\<^sub>s\<^sub>s\<^sub>t_def by auto
      thus ?thesis using Cons Receive by simp
    qed (auto simp add: ik\<^sub>s\<^sub>s\<^sub>t_def)
  qed (simp_all add: ik\<^sub>s\<^sub>s\<^sub>t_def strand_sem_ik_mono_stateful)

  show ?thesis by (metis 1 2 3)
qed

section {* Small lemmas *}
lemma subst_pair_compose[simp]: "d \<cdot>p (\<delta> \<circ>\<^sub>s \<I>) = d \<cdot>p \<delta> \<cdot>p \<I>"
proof -
  obtain t s where "d = (t,s)" by moura
  thus ?thesis by auto
qed

lemma to_pair_eq[dest]: "to_pair d = to_pair d' \<Longrightarrow> d = d'"
proof -
  obtain t s t' s' where "d = (t,s)" "d' = (t',s')" by moura
  thus "to_pair d = to_pair d' \<Longrightarrow> d = d'" by simp
qed

lemma to_pair_subst: "to_pair d \<cdot> \<delta> = to_pair (d \<cdot>p \<delta>)"
using surj_pair[of d] by force  

lemma to_pair_eq_subst: "to_pair d \<cdot> \<delta> = to_pair d' \<cdot> \<theta> \<Longrightarrow> d \<cdot>p \<delta> = d' \<cdot>p \<theta>"
by (metis to_pair_subst to_pair_eq[of "d \<cdot>p \<delta>" "d' \<cdot>p \<theta>"])

lemma sst_vars_disj_cons[simp]: "FV\<^sub>s\<^sub>s\<^sub>t (a#A) \<inter> bvars\<^sub>s\<^sub>s\<^sub>t (a#A) = {} \<Longrightarrow> FV\<^sub>s\<^sub>s\<^sub>t A \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}"
unfolding FV\<^sub>s\<^sub>s\<^sub>t_def bvars\<^sub>s\<^sub>s\<^sub>t_def by auto

lemma setops\<^sub>s\<^sub>s\<^sub>t_cons_subset[simp]: "setops\<^sub>s\<^sub>s\<^sub>t A \<subseteq> setops\<^sub>s\<^sub>s\<^sub>t (a#A)"
proof (induction A arbitrary: a)
  case (Cons a A) thus ?case by (cases a) auto
qed simp

lemma setops\<^sub>s\<^sub>s\<^sub>t_append: "setops\<^sub>s\<^sub>s\<^sub>t (A@B) = setops\<^sub>s\<^sub>s\<^sub>t A \<union> setops\<^sub>s\<^sub>s\<^sub>t B"
proof (induction A)
  case (Cons a A) thus ?case by (cases a) auto
qed simp

lemma setops\<^sub>s\<^sub>s\<^sub>t_are_pairs: "t \<in> setops\<^sub>s\<^sub>s\<^sub>t A \<Longrightarrow> \<exists>s s'. t = to_pair (s,s')"
by (induct A rule: setops\<^sub>s\<^sub>s\<^sub>t.induct) auto

lemma to_pair_wf\<^sub>t\<^sub>r\<^sub>m: "wf\<^sub>t\<^sub>r\<^sub>m t \<Longrightarrow> wf\<^sub>t\<^sub>r\<^sub>m t' \<Longrightarrow> wf\<^sub>t\<^sub>r\<^sub>m (to_pair (t,t'))"
using pair_arity unfolding wf\<^sub>t\<^sub>r\<^sub>m_def by auto

lemma sublists_set_subset:
  assumes "Di \<in> set (sublists D)"
  shows "set Di \<subseteq> set D" "set [d\<leftarrow>D. d \<notin> set Di] \<subseteq> set D"
proof -
  show "set Di \<subseteq> set D" using assms sublists_powset[of D] by auto
  show "set [d\<leftarrow>D. d \<notin> set Di] \<subseteq> set D" by auto
qed

lemma subset_subst_pairs_diff_exists:
  fixes \<I>::"('a,'b) subst" and D D'::"('a,'b) dbstate"
  shows "\<exists>Di. Di \<subseteq> D \<and> Di \<cdot>pairs \<I> = (D \<cdot>pairs \<I>) - D'"
by (metis (no_types, lifting) Diff_subset subset_image_iff)

lemma subset_subst_pairs_diff_exists':
  fixes \<I>::"('a,'b) subst" and D::"('a,'b) dbstate"
  assumes "finite D"
  shows "\<exists>Di. Di \<subseteq> D \<and> Di \<cdot>pairs \<I> \<subseteq> {d \<cdot>p \<I>} \<and> d \<cdot>p \<I> \<notin> (D - Di) \<cdot>pairs \<I>"
using assms
proof (induction D rule: finite_induct)
  case (insert d' D)
  then obtain Di where IH: "Di \<subseteq> D" "Di \<cdot>pairs \<I> \<subseteq> {d \<cdot>p \<I>}" "d \<cdot>p \<I> \<notin> (D - Di) \<cdot>pairs \<I>" by moura
  show ?case
  proof (cases "d' \<cdot>p \<I> = d \<cdot>p \<I>")
    case True
    hence "insert d' Di \<subseteq> insert d' D" "insert d' Di \<cdot>pairs \<I> \<subseteq> {d \<cdot>p \<I>}"
          "d \<cdot>p \<I> \<notin> (insert d' D - insert d' Di) \<cdot>pairs \<I>" 
      using IH by auto
    thus ?thesis by metis
  next
    case False
    hence "Di \<subseteq> insert d' D" "Di \<cdot>pairs \<I> \<subseteq> {d \<cdot>p \<I>}"
          "d \<cdot>p \<I> \<notin> (insert d' D - Di) \<cdot>pairs \<I>" 
      using IH by auto
    thus ?thesis by metis
  qed
qed simp

lemma subset_sublist_exists:
  "Di \<subseteq> set D \<Longrightarrow> \<exists>Di'. set Di' = Di \<and> Di' \<in> set (sublists D)"
proof (induction D arbitrary: Di)
  case Cons thus ?case by (metis (no_types, lifting) Pow_iff imageE sublists_powset) 
qed simp

lemma trms\<^sub>s\<^sub>s\<^sub>t_cons: "trms\<^sub>s\<^sub>s\<^sub>t (a#A) = trms\<^sub>s\<^sub>s\<^sub>t [a] \<union> trms\<^sub>s\<^sub>s\<^sub>t A"
unfolding trms\<^sub>s\<^sub>s\<^sub>t_def by force

lemma trms\<^sub>s\<^sub>s\<^sub>t_append: "trms\<^sub>s\<^sub>s\<^sub>t (A@B) = trms\<^sub>s\<^sub>s\<^sub>t A \<union> trms\<^sub>s\<^sub>s\<^sub>t B"
unfolding trms\<^sub>s\<^sub>s\<^sub>t_def by force

lemma tfr\<^sub>s\<^sub>s\<^sub>t_append: "tfr\<^sub>s\<^sub>s\<^sub>t (A@B) \<Longrightarrow> tfr\<^sub>s\<^sub>s\<^sub>t A"
proof -
  assume assms: "tfr\<^sub>s\<^sub>s\<^sub>t (A@B)"
  let ?M = "trms\<^sub>s\<^sub>s\<^sub>t A \<union> setops\<^sub>s\<^sub>s\<^sub>t A"
  let ?N = "trms\<^sub>s\<^sub>s\<^sub>t (A@B) \<union> setops\<^sub>s\<^sub>s\<^sub>t (A@B)"
  let ?P = "\<lambda>t t'. \<forall>x \<in> FV t \<union> FV t'. \<exists>a. \<Gamma> (Var x) = Var a"
  let ?Q = "\<lambda>X t t'. \<forall>x \<in> (FV t \<union> FV t')-set X. \<exists>a. \<Gamma> (Var x) = Var a"
  have *: "SMP\<^sub>s\<^sub>t ?M - Var`\<V> \<subseteq> SMP\<^sub>s\<^sub>t ?N - Var`\<V>" "?M \<subseteq> ?N"
    using SMP_mono[of ?M ?N] setops\<^sub>s\<^sub>s\<^sub>t_append[of A B]
    by auto
  { fix s t assume **: "tfr\<^sub>s\<^sub>t ?N" "s \<in> SMP\<^sub>s\<^sub>t ?M - Var`\<V>" "t \<in> SMP\<^sub>s\<^sub>t ?M - Var`\<V>" "(\<exists>\<delta>. Unifier \<delta> s t)"
    hence "s \<in> SMP\<^sub>s\<^sub>t ?N - Var`\<V>" "t \<in> SMP\<^sub>s\<^sub>t ?N - Var`\<V>" using * by auto
    hence "\<Gamma> s = \<Gamma> t" using **(1,4) by blast
  } moreover have "\<forall>t \<in> ?N. wf\<^sub>t\<^sub>r\<^sub>m t \<Longrightarrow> \<forall>t \<in> ?M. wf\<^sub>t\<^sub>r\<^sub>m t" using * by blast
  ultimately have "tfr\<^sub>s\<^sub>t ?N \<Longrightarrow> tfr\<^sub>s\<^sub>t ?M" by blast
  hence "tfr\<^sub>s\<^sub>t ?M" using assms unfolding tfr\<^sub>s\<^sub>s\<^sub>t_def by metis
  moreover have
      "\<forall>t t'. Equality t t' \<in> set (A@B) \<and> (\<exists>\<delta>. Unifier \<delta> t t') \<longrightarrow> \<Gamma> t = \<Gamma> t'"
      "\<forall>X t t'. Inequality X t t' \<in> set (A@B) \<longrightarrow> ?Q X t t'"
      "\<forall>t t'. Insert t t' \<in> set (A@B) \<longrightarrow> ?P t t'"
      "\<forall>t t'. Delete t t' \<in> set (A@B) \<longrightarrow> ?P t t'"
      "\<forall>X t t'. NotInSet X t t' \<in> set (A@B) \<longrightarrow> ?Q X t t'"
    using assms unfolding tfr\<^sub>s\<^sub>s\<^sub>t_def eqs_tfr\<^sub>s\<^sub>s\<^sub>t_def by metis+
  hence "\<forall>t t'. Equality t t' \<in> set A \<and> (\<exists>\<delta>. Unifier \<delta> t t') \<longrightarrow> \<Gamma> t = \<Gamma> t'"
        "\<forall>X t t'. Inequality X t t' \<in> set A \<longrightarrow> ?Q X t t'"
        "\<forall>t t'. Insert t t' \<in> set A \<longrightarrow> ?P t t'"
        "\<forall>t t'. Delete t t' \<in> set A \<longrightarrow> ?P t t'"
        "\<forall>X t t'. NotInSet X t t' \<in> set A \<longrightarrow> ?Q X t t'"
    by auto
  ultimately show "tfr\<^sub>s\<^sub>s\<^sub>t A" unfolding tfr\<^sub>s\<^sub>s\<^sub>t_def eqs_tfr\<^sub>s\<^sub>s\<^sub>t_def by blast
qed

lemma tfr\<^sub>s\<^sub>s\<^sub>t_append': "tfr\<^sub>s\<^sub>s\<^sub>t (A@B) \<Longrightarrow> tfr\<^sub>s\<^sub>s\<^sub>t B"
proof -
  assume assms: "tfr\<^sub>s\<^sub>s\<^sub>t (A@B)"
  let ?M = "trms\<^sub>s\<^sub>s\<^sub>t B \<union> setops\<^sub>s\<^sub>s\<^sub>t B"
  let ?N = "trms\<^sub>s\<^sub>s\<^sub>t (A@B) \<union> setops\<^sub>s\<^sub>s\<^sub>t (A@B)"
  let ?P = "\<lambda>t t'. \<forall>x \<in> FV t \<union> FV t'. \<exists>a. \<Gamma> (Var x) = Var a"
  let ?Q = "\<lambda>X t t'. \<forall>x \<in> (FV t \<union> FV t')-set X. \<exists>a. \<Gamma> (Var x) = Var a"
  have *: "SMP\<^sub>s\<^sub>t ?M - Var`\<V> \<subseteq> SMP\<^sub>s\<^sub>t ?N - Var`\<V>" "?M \<subseteq> ?N"
    using SMP_mono[of ?M ?N] setops\<^sub>s\<^sub>s\<^sub>t_append[of A B]
    by auto
  { fix s t assume **: "tfr\<^sub>s\<^sub>t ?N" "s \<in> SMP\<^sub>s\<^sub>t ?M - Var`\<V>" "t \<in> SMP\<^sub>s\<^sub>t ?M - Var`\<V>" "(\<exists>\<delta>. Unifier \<delta> s t)"
    hence "s \<in> SMP\<^sub>s\<^sub>t ?N - Var`\<V>" "t \<in> SMP\<^sub>s\<^sub>t ?N - Var`\<V>" using * by auto
    hence "\<Gamma> s = \<Gamma> t" using **(1,4) by blast
  } moreover have "\<forall>t \<in> ?N. wf\<^sub>t\<^sub>r\<^sub>m t \<Longrightarrow> \<forall>t \<in> ?M. wf\<^sub>t\<^sub>r\<^sub>m t" using * by blast
  ultimately have "tfr\<^sub>s\<^sub>t ?N \<Longrightarrow> tfr\<^sub>s\<^sub>t ?M" by blast
  hence "tfr\<^sub>s\<^sub>t ?M" using assms unfolding tfr\<^sub>s\<^sub>s\<^sub>t_def by metis
  moreover have
      "\<forall>t t'. Equality t t' \<in> set (A@B) \<and> (\<exists>\<delta>. Unifier \<delta> t t') \<longrightarrow> \<Gamma> t = \<Gamma> t'"
      "\<forall>X t t'. Inequality X t t' \<in> set (A@B) \<longrightarrow> ?Q X t t'"
      "\<forall>t t'. Insert t t' \<in> set (A@B) \<longrightarrow> ?P t t'"
      "\<forall>t t'. Delete t t' \<in> set (A@B) \<longrightarrow> ?P t t'"
      "\<forall>X t t'. NotInSet X t t' \<in> set (A@B) \<longrightarrow> ?Q X t t'"
    using assms unfolding tfr\<^sub>s\<^sub>s\<^sub>t_def eqs_tfr\<^sub>s\<^sub>s\<^sub>t_def by metis+
  hence "\<forall>t t'. Equality t t' \<in> set B \<and> (\<exists>\<delta>. Unifier \<delta> t t') \<longrightarrow> \<Gamma> t = \<Gamma> t'"
        "\<forall>X t t'. Inequality X t t' \<in> set B \<longrightarrow> ?Q X t t'"
        "\<forall>t t'. Insert t t' \<in> set B \<longrightarrow> ?P t t'"
        "\<forall>t t'. Delete t t' \<in> set B \<longrightarrow> ?P t t'"
        "\<forall>X t t'. NotInSet X t t' \<in> set B \<longrightarrow> ?Q X t t'"
    by auto
  ultimately show "tfr\<^sub>s\<^sub>s\<^sub>t B" unfolding tfr\<^sub>s\<^sub>s\<^sub>t_def eqs_tfr\<^sub>s\<^sub>s\<^sub>t_def by blast
qed

lemma tfr\<^sub>s\<^sub>s\<^sub>t_cons: "tfr\<^sub>s\<^sub>s\<^sub>t (a#A) \<Longrightarrow> tfr\<^sub>s\<^sub>s\<^sub>t A"
using tfr\<^sub>s\<^sub>s\<^sub>t_append'[of "[a]" A] by simp

lemma tr_finite: "finite (tr A D)"
by (induct A D rule: tr.induct) auto

lemma tr_trms_subset:
  "A' \<in> tr A D \<Longrightarrow> trms\<^sub>s\<^sub>t A' \<subseteq> trms\<^sub>s\<^sub>s\<^sub>t A \<union> setops\<^sub>s\<^sub>s\<^sub>t A \<union> to_pair ` set D"
proof (induction A D arbitrary: A' rule: tr.induct)
  case 1 thus ?case by simp
next
  case (2 t A D)
  then obtain A'' where A'': "A' = snd\<langle>t\<rangle>#A''" "A'' \<in> tr A D" by moura
  hence "trms\<^sub>s\<^sub>t A'' \<subseteq> trms\<^sub>s\<^sub>s\<^sub>t A \<union> setops\<^sub>s\<^sub>s\<^sub>t A \<union> to_pair ` set D" by (metis "2.IH")
  thus ?case using A'' by auto
next
  case (3 t A D)
  then obtain A'' where A'': "A' = rcv\<langle>t\<rangle>#A''" "A'' \<in> tr A D" by moura
  hence "trms\<^sub>s\<^sub>t A'' \<subseteq> trms\<^sub>s\<^sub>s\<^sub>t A \<union> setops\<^sub>s\<^sub>s\<^sub>t A \<union> to_pair ` set D" by (metis "3.IH")
  thus ?case using A'' by auto
next
  case (4 t t' A D)
  then obtain A'' where A'': "A' = t \<doteq> t'#A''" "A'' \<in> tr A D" by moura
  hence "trms\<^sub>s\<^sub>t A'' \<subseteq> trms\<^sub>s\<^sub>s\<^sub>t A \<union> setops\<^sub>s\<^sub>s\<^sub>t A \<union> to_pair ` set D" by (metis "4.IH")
  thus ?case using A'' by auto
next
  case (5 X t t' A D)
  then obtain A'' where A'': "A' = \<nexists>X\<langle>t \<doteq> t'\<rangle>#A''" "A'' \<in> tr A D" by moura
  hence "trms\<^sub>s\<^sub>t A'' \<subseteq> trms\<^sub>s\<^sub>s\<^sub>t A \<union> setops\<^sub>s\<^sub>s\<^sub>t A \<union> to_pair ` set D" by (metis "5.IH")
  thus ?case using A'' by auto
next
  case (6 t s A D)
  hence "A' \<in> tr A (List.insert (t,s) D)" by simp
  hence "trms\<^sub>s\<^sub>t A' \<subseteq> trms\<^sub>s\<^sub>s\<^sub>t A \<union> setops\<^sub>s\<^sub>s\<^sub>t A \<union> to_pair ` set (List.insert (t, s) D)"
    by (metis "6.IH")
  thus ?case by auto
next
  case (7 t s A D)
  from 7 obtain Di A'' B C where A'':
      "Di \<in> set (sublists D)" "A'' \<in> tr A [d\<leftarrow>D. d \<notin> set Di]" "A' = (B@C)@A''"
      "B = map (\<lambda>d. to_pair (t,s) \<doteq> to_pair d) Di"
      "C = map (\<lambda>d. \<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>) [d\<leftarrow>D. d \<notin> set Di]"
    by moura
  hence "trms\<^sub>s\<^sub>t A'' \<subseteq> trms\<^sub>s\<^sub>s\<^sub>t A \<union> setops\<^sub>s\<^sub>s\<^sub>t A \<union> to_pair ` set [d\<leftarrow>D. d \<notin> set Di]"
    by (metis "7.IH")
  hence "trms\<^sub>s\<^sub>t A'' \<subseteq> trms\<^sub>s\<^sub>s\<^sub>t (Delete t s#A) \<union> setops\<^sub>s\<^sub>s\<^sub>t (Delete t s#A) \<union> to_pair ` set D"
    by auto
  moreover have "trms\<^sub>s\<^sub>t (B@C) \<subseteq> insert (to_pair (t,s)) (to_pair ` set D)"
    using A''(4,5) sublists_set_subset[OF A''(1)] by force
  moreover have "to_pair (t,s) \<in> setops\<^sub>s\<^sub>s\<^sub>t (Delete t s#A)" by simp
  ultimately show ?case using A''(3) trms\<^sub>s\<^sub>t_append[of "B@C" A'] by auto
next
  case (8 t s A D)
  from 8 obtain d A'' where A'': "d \<in> set D" "A'' \<in> tr A D" "A' = to_pair (t,s) \<doteq> to_pair d#A''"
    by moura
  hence "trms\<^sub>s\<^sub>t A'' \<subseteq> trms\<^sub>s\<^sub>s\<^sub>t A \<union> setops\<^sub>s\<^sub>s\<^sub>t A \<union> to_pair ` set D" by (metis "8.IH")
  moreover have "trms\<^sub>s\<^sub>t A' = {to_pair (t,s), to_pair d} \<union> trms\<^sub>s\<^sub>t A''"
    using A''(1,3) by auto
  ultimately show ?case using A''(1) by auto
next
  case (9 X t s A D)
  from 9 obtain A'' where A'':
      "A'' \<in> tr A D" "A' = (map (\<lambda>d. \<nexists>X\<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>) D)@A''"
    by moura
  hence "trms\<^sub>s\<^sub>t A'' \<subseteq> trms\<^sub>s\<^sub>s\<^sub>t A \<union> setops\<^sub>s\<^sub>s\<^sub>t A \<union> to_pair ` set D" by (metis "9.IH")
  hence "trms\<^sub>s\<^sub>t A' \<subseteq> {to_pair (t,s)} \<union> trms\<^sub>s\<^sub>s\<^sub>t A \<union> setops\<^sub>s\<^sub>s\<^sub>t A \<union> to_pair ` set D" using A'' by auto
  thus ?case by auto
qed

lemma tr_vars_subset:
  assumes "A' \<in> tr A D"
  shows "FV\<^sub>s\<^sub>t A' \<subseteq> FV\<^sub>s\<^sub>s\<^sub>t A \<union> (\<Union>(t,t') \<in> set D. FV t \<union> FV t')" (is ?P)
  and "bvars\<^sub>s\<^sub>t A' \<subseteq> bvars\<^sub>s\<^sub>s\<^sub>t A" (is ?Q)
proof -
  show ?P using assms
  proof (induction A arbitrary: A' D rule: strand_sem_stateful_induct)
    case (ConsIn A' D t s A)
    then obtain A'' d where *:
        "d \<in> set D" "A' = to_pair (t,s) \<doteq> to_pair d#A''" "A'' \<in> tr A D"
      by moura
    hence "FV\<^sub>s\<^sub>t A'' \<subseteq> FV\<^sub>s\<^sub>s\<^sub>t A \<union> (\<Union>(t,t')\<in>set D. FV t \<union> FV t')" by (metis ConsIn.IH)
    thus ?case using * by auto
  next
    case (ConsDel A' D t s A)
    let ?DFV = "\<lambda>D. (\<Union>(t,t')\<in>set D. FV t \<union> FV t')"
    let ?fltD = "\<lambda>Di. filter (\<lambda>d. d \<notin> set Di) D"
    let ?constr = "\<lambda>Di. (map (\<lambda>d. to_pair (t,s) \<doteq> to_pair d) Di)@
                        (map (\<lambda>d. \<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>) (?fltD Di))"
    from ConsDel obtain A'' Di where *:
        "Di \<in> set (sublists D)" "A' = (?constr Di)@A''" "A'' \<in> tr A (?fltD Di)"
      by moura
    hence "FV\<^sub>s\<^sub>t A'' \<subseteq> FV\<^sub>s\<^sub>s\<^sub>t A \<union> ?DFV (filter (\<lambda>d. d \<notin> set Di) D)" by (metis ConsDel.IH)
    moreover have "?DFV (?fltD Di) \<subseteq> ?DFV D" by auto
    moreover have "?DFV Di \<subseteq> ?DFV D" using sublists_set_subset(1)[OF *(1)] by fast
    moreover have "FV\<^sub>s\<^sub>t (?constr Di) \<subseteq> FV t \<union> FV s \<union> ?DFV Di \<union> ?DFV (?fltD Di)" by auto
    moreover have "FV\<^sub>s\<^sub>s\<^sub>t (Delete t s # A) = FV t \<union> FV s \<union> FV\<^sub>s\<^sub>s\<^sub>t A" by auto
    moreover have "FV\<^sub>s\<^sub>t A' = FV\<^sub>s\<^sub>t (?constr Di) \<union> FV\<^sub>s\<^sub>t A''" using * by force
    ultimately show ?case by blast
  qed fastforce+

  show ?Q using assms by (induct A arbitrary: A' D rule: strand_sem_stateful_induct) fastforce+
qed

lemma tr_vars_disj:
  assumes "A' \<in> tr A D" "\<forall>(t,t') \<in> set D. (FV t \<union> FV t') \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}"
  and "FV\<^sub>s\<^sub>s\<^sub>t A \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}"
  shows "FV\<^sub>s\<^sub>t A' \<inter> bvars\<^sub>s\<^sub>t A' = {}"
using assms tr_vars_subset by fast

lemma wf_to_pair_ineqs_map:
  assumes "wf\<^sub>s\<^sub>t X A"
  shows "wf\<^sub>s\<^sub>t X (map (\<lambda>d. \<nexists>Y\<langle>to_pair (t, s) \<doteq> to_pair d\<rangle>) D@A)"
using assms by (induct D) auto

lemma wf_to_pair_eqs_ineqs_map:
  fixes A::"('fun,'var) strand"
  assumes "wf\<^sub>s\<^sub>t X A" "Di \<in> set (sublists D)" "\<forall>(t,t') \<in> set D. FV t \<union> FV t' \<subseteq> X"
  shows "wf\<^sub>s\<^sub>t X ((map (\<lambda>d. to_pair (t,s) \<doteq> to_pair d) Di)@
                 (map (\<lambda>d. \<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>) (filter (\<lambda>d. d \<notin> set Di) D))@A)"
proof -
  let ?c1 = "map (\<lambda>d. to_pair (t,s) \<doteq> to_pair d) Di"
  let ?c2 = "map (\<lambda>d. \<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>) (filter (\<lambda>d. d \<notin> set Di) D)"
  have 1: "wf\<^sub>s\<^sub>t X (?c2@A)" using wf_to_pair_ineqs_map[OF assms(1)] by simp
  have 2: "\<forall>(t,t') \<in> set Di. FV t \<union> FV t' \<subseteq> X" 
    using assms(2,3) by (meson contra_subsetD sublists_set_subset(1))
  { fix B::"('fun,'var) strand" assume "wf\<^sub>s\<^sub>t X B" hence "wf\<^sub>s\<^sub>t X (?c1@B)" using 2 by (induct Di) auto
  } thus ?thesis using 1 by simp
qed

section {* The typing result *}

lemma tr_wf:
  assumes "\<forall>(t,t') \<in> set D. (FV t \<union> FV t') \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}"
  and "\<forall>(t,t') \<in> set D. FV t \<union> FV t' \<subseteq> X"
  and "wf\<^sub>s\<^sub>s\<^sub>t X A" "FV\<^sub>s\<^sub>s\<^sub>t A \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}"
  and "A' \<in> tr A D"
  shows "wf\<^sub>s\<^sub>t X A'"
proof -
  def P \<equiv> "\<lambda>(D::('fun,'var) dbstatelist) (A::('fun,'var) statefulstrand).
                (\<forall>(t,t') \<in> set D. (FV t \<union> FV t') \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}) \<and> FV\<^sub>s\<^sub>s\<^sub>t A \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}"
  have "P D A" using assms(1,4) by (simp add: P_def)
  with assms(5,3,2) show ?thesis
  proof (induction A arbitrary: A' D X rule: wf\<^sub>s\<^sub>s\<^sub>t.induct)
    case 1 thus ?case by simp
  next
    case (2 X t A A')
    then obtain A'' where A'': "A' = rcv\<langle>t\<rangle>#A''" "A'' \<in> tr A D" "FV t \<subseteq> X" by moura
    have *: "wf\<^sub>s\<^sub>s\<^sub>t X A" "\<forall>(s,s') \<in> set D. FV s \<union> FV s' \<subseteq> X" "P D A"
      using 2 unfolding P_def by auto
    show ?case using "2.IH"[OF A''(2) *] A''(1,3) by simp
  next
    case (3 X t A A')
    then obtain A'' where A'': "A' = snd\<langle>t\<rangle>#A''" "A'' \<in> tr A D" by moura
    have *: "wf\<^sub>s\<^sub>s\<^sub>t (X \<union> FV t) A" "\<forall>(s,s') \<in> set D. FV s \<union> FV s' \<subseteq> X \<union> FV t" "P D A"
      using 3 unfolding P_def by auto
    show ?case using "3.IH"[OF A''(2) *] A''(1) by simp
  next
    case (4 X t t' A A')
    then obtain A'' where A'': "A' = t \<doteq> t'#A''" "A'' \<in> tr A D" "FV t' \<subseteq> X" by moura
    have *: "wf\<^sub>s\<^sub>s\<^sub>t (X \<union> FV t) A" "\<forall>(s,s') \<in> set D. FV s \<union> FV s' \<subseteq> X \<union> FV t" "P D A"
      using 4 unfolding P_def by auto
    show ?case using "4.IH"[OF A''(2) *] A''(1,3) by simp
  next
    case (5 X Y t t' A A')
    then obtain A'' where A'': "A' = \<nexists>Y\<langle>t \<doteq> t'\<rangle>#A''" "A'' \<in> tr A D" by moura
    have *: "wf\<^sub>s\<^sub>s\<^sub>t X A" "\<forall>(s,s') \<in> set D. FV s \<union> FV s' \<subseteq> X" using 5 by auto
    have "bvars\<^sub>s\<^sub>s\<^sub>t A \<subseteq> bvars\<^sub>s\<^sub>s\<^sub>t (Inequality Y t t'#A)" "FV\<^sub>s\<^sub>s\<^sub>t A \<subseteq> FV\<^sub>s\<^sub>s\<^sub>t (Inequality Y t t'#A)"
      by auto
    hence **:  "P D A" using 5 unfolding P_def by blast
    show ?case using "5.IH"[OF A''(2) * **] A''(1) by simp
  next
    case (6 X t s A A')
    hence A': "A' \<in> tr A (List.insert (t,s) D)" "FV t \<subseteq> X" "FV s \<subseteq> X" by auto
    have *: "wf\<^sub>s\<^sub>s\<^sub>t X A" "\<forall>(s,s') \<in> set (List.insert (t,s) D). FV s \<union> FV s' \<subseteq> X" using 6 by auto
    have **: "P (List.insert (t,s) D) A" using 6 unfolding P_def by auto
    show ?case using "6.IH"[OF A'(1) * **] A'(2,3) by simp
  next
    case (7 X t s A A')
    let ?constr = "\<lambda>Di. (map (\<lambda>d. to_pair (t,s) \<doteq> to_pair d) Di)@
                        (map (\<lambda>d. \<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>) (filter (\<lambda>d. d \<notin> set Di) D))"
    from 7 obtain Di A'' where A'':
        "A' = ?constr Di@A''" "A'' \<in> tr A (filter (\<lambda>d. d \<notin> set Di) D)"
        "Di \<in> set (sublists D)"
      by moura
    have *: "wf\<^sub>s\<^sub>s\<^sub>t X A" "\<forall>(t',s') \<in> set (filter (\<lambda>d. d \<notin> set Di) D). FV t' \<union> FV s' \<subseteq> X"
      using 7 by auto
    have **: "P (filter (\<lambda>d. d \<notin> set Di) D) A" using 7 unfolding P_def by auto
    have ***: "\<forall>(t, t') \<in> set D. FV t \<union> FV t' \<subseteq> X" using 7 by auto
    show ?case using "7.IH"[OF A''(2) * **] A''(1) wf_to_pair_eqs_ineqs_map[OF _ A''(3) ***] by simp
  next
    case (8 X t s A A')
    then obtain d A'' where A'': "A' = to_pair (t,s) \<doteq> to_pair d#A''" "A'' \<in> tr A D" "d \<in> set D"
      by moura
    have *: "wf\<^sub>s\<^sub>s\<^sub>t (X \<union> FV t \<union> FV s) A" "\<forall>(t',s')\<in>set D. FV t' \<union> FV s' \<subseteq> X \<union> FV t \<union> FV s" "P D A"
      using 8 unfolding P_def by auto
    have **: "FV (to_pair d) \<subseteq> X" using A''(3) "8.prems"(3) by fastforce
    have ***: "FV (to_pair (t,s)) = FV s \<union> FV t" by auto
    show ?case using "8.IH"[OF A''(2) *] A''(1) ** *** by (simp add: Un_assoc)
  next
    case (9 X Y t s A A')
    then obtain A'' where A'': "A' = (map (\<lambda>d. \<nexists>Y\<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>) D)@A''" "A'' \<in> tr A D"
      by moura
    have *: "wf\<^sub>s\<^sub>s\<^sub>t X A" "\<forall>(t',s') \<in> set D. FV t' \<union> FV s' \<subseteq> X" using 9 by auto
    have "bvars\<^sub>s\<^sub>s\<^sub>t A \<subseteq> bvars\<^sub>s\<^sub>s\<^sub>t (NotInSet Y t s#A)" "FV\<^sub>s\<^sub>s\<^sub>t A \<subseteq> FV\<^sub>s\<^sub>s\<^sub>t (NotInSet Y t s#A)"
      by auto
    hence **:  "P D A" using 9 unfolding P_def by blast
    show ?case using "9.IH"[OF A''(2) * **] A''(1) wf_to_pair_ineqs_map by simp
  qed
qed

lemma setops_wf_trm: "\<forall>t \<in> trms\<^sub>s\<^sub>s\<^sub>t A. wf\<^sub>t\<^sub>r\<^sub>m t \<Longrightarrow> \<forall>t \<in> setops\<^sub>s\<^sub>s\<^sub>t A. wf\<^sub>t\<^sub>r\<^sub>m t"
using to_pair_wf\<^sub>t\<^sub>r\<^sub>m by (induct A rule: setops\<^sub>s\<^sub>s\<^sub>t.induct) auto

lemma tr_eqs_tfr:
  assumes "A' \<in> tr A D" "eqs_tfr\<^sub>s\<^sub>s\<^sub>t A"
  and "\<forall>(t,s) \<in> set D. \<forall>x \<in> FV t \<union> FV s. \<exists>a. \<Gamma> (Var x) = TAtom a"
  and "\<forall>t \<in> setops\<^sub>s\<^sub>s\<^sub>t A \<union> to_pair ` set D. \<forall>t' \<in> setops\<^sub>s\<^sub>s\<^sub>t A \<union> to_pair ` set D.
          (\<exists>\<delta>. Unifier \<delta> t t') \<longrightarrow> \<Gamma> t = \<Gamma> t'" (is "?P A D")
  shows "eqs_tfr\<^sub>s\<^sub>t A'"
using assms
proof (induction A D arbitrary: A' rule: tr.induct)
  case 1 thus ?case by simp
next
  case (2 t A D)
  from "2.prems"(1) obtain A'' where A'': "A' = snd\<langle>t\<rangle>#A''" "A'' \<in> tr A D" by moura
  have "eqs_tfr\<^sub>s\<^sub>s\<^sub>t A" using "2.prems"(2) unfolding eqs_tfr\<^sub>s\<^sub>s\<^sub>t_def by simp
  moreover have "?P A D" using "2.prems"(4) setops\<^sub>s\<^sub>s\<^sub>t_cons_subset[of A] by fast
  ultimately have "eqs_tfr\<^sub>s\<^sub>t A''" using "2.IH"[OF A''(2) _ "2.prems"(3)] by metis
  thus ?case using A''(1) by simp
next
  case (3 t A D)
  from "3.prems"(1) obtain A'' where A'': "A' = rcv\<langle>t\<rangle>#A''" "A'' \<in> tr A D" by moura
  have "eqs_tfr\<^sub>s\<^sub>s\<^sub>t A" using "3.prems"(2) unfolding eqs_tfr\<^sub>s\<^sub>s\<^sub>t_def by simp
  moreover have "?P A D" using "3.prems"(4) setops\<^sub>s\<^sub>s\<^sub>t_cons_subset[of A] by fast
  ultimately have "eqs_tfr\<^sub>s\<^sub>t A''" using "3.IH"[OF A''(2) _ "3.prems"(3)] by metis
  thus ?case using A''(1) by simp
next
  case (4 t t' A D)
  from "4.prems"(1) obtain A'' where A'': "A' = t \<doteq> t'#A''" "A'' \<in> tr A D" by moura
  have "eqs_tfr\<^sub>s\<^sub>s\<^sub>t A" using "4.prems"(2) unfolding eqs_tfr\<^sub>s\<^sub>s\<^sub>t_def by simp
  moreover have "?P A D" using "4.prems"(4) setops\<^sub>s\<^sub>s\<^sub>t_cons_subset[of A] by fast
  ultimately have "eqs_tfr\<^sub>s\<^sub>t A''" using "4.IH"[OF A''(2) _ "4.prems"(3)] by metis
  moreover have "(\<exists>\<delta>. Unifier \<delta> t t') \<Longrightarrow> \<Gamma> t = \<Gamma> t'"
    using "4.prems"(2) unfolding eqs_tfr\<^sub>s\<^sub>s\<^sub>t_def by simp
  ultimately show ?case using A''(1) by auto
next
  case (5 X t t' A D)
  from "5.prems"(1) obtain A'' where A'': "A' = \<nexists>X\<langle>t \<doteq> t'\<rangle>#A''" "A'' \<in> tr A D" by moura
  have "eqs_tfr\<^sub>s\<^sub>s\<^sub>t A" using "5.prems"(2) unfolding eqs_tfr\<^sub>s\<^sub>s\<^sub>t_def by simp
  moreover have "?P A D" using "5.prems"(4) setops\<^sub>s\<^sub>s\<^sub>t_cons_subset[of A] by fast
  ultimately have "eqs_tfr\<^sub>s\<^sub>t A''" using "5.IH"[OF A''(2) _ "5.prems"(3)] by metis
  moreover have "\<forall>x \<in> (FV t \<union> FV t')-set X. \<exists>a. \<Gamma> (Var x) = TAtom a"
    using "5.prems"(2) unfolding eqs_tfr\<^sub>s\<^sub>s\<^sub>t_def by simp
  ultimately show ?case using A''(1) by auto
next
  case (6 t s A D)
  from "6.prems"(1) have A': "A' \<in> tr A (List.insert (t,s) D)" by simp
  have *: "eqs_tfr\<^sub>s\<^sub>s\<^sub>t A" using "6.prems"(2,3) unfolding eqs_tfr\<^sub>s\<^sub>s\<^sub>t_def by simp
  have "\<forall>x \<in> FV t \<union> FV s. \<exists>a. \<Gamma> (Var x) = TAtom a"
    using "6.prems"(2) unfolding eqs_tfr\<^sub>s\<^sub>s\<^sub>t_def by simp
  hence **: "\<forall>(t,s) \<in> set (List.insert (t,s) D). \<forall>x \<in> FV t \<union> FV s. \<exists>a. \<Gamma> (Var x) = TAtom a"
    using "6.prems"(3) by simp
  have "setops\<^sub>s\<^sub>s\<^sub>t (Insert t s#A) \<union> to_pair`set D = setops\<^sub>s\<^sub>s\<^sub>t A \<union> to_pair`set (List.insert (t,s) D)"
    by auto
  hence ***: "?P A (List.insert (t,s) D)" using "6.prems"(4) by metis
  have "eqs_tfr\<^sub>s\<^sub>t A'" using "6.IH"[OF A' * ** ***] by metis
  thus ?case using A'(1) by auto
next
  case (7 t s A D)
  let ?constr = "\<lambda>Di. (map (\<lambda>d. to_pair (t,s) \<doteq> to_pair d) Di)@
                      (map (\<lambda>d. \<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>) (filter (\<lambda>d. d \<notin> set Di) D))"
  from 7 obtain Di A'' where A'':
      "A' = ?constr Di@A''" "A'' \<in> tr A (filter (\<lambda>d. d \<notin> set Di) D)"
      "Di \<in> set (sublists D)"
    by moura
  have *: "eqs_tfr\<^sub>s\<^sub>s\<^sub>t A" using "7.prems"(2,3) unfolding eqs_tfr\<^sub>s\<^sub>s\<^sub>t_def by simp
  have "set (filter (\<lambda>d. d \<notin> set Di) D) \<subseteq> set D"
       "setops\<^sub>s\<^sub>s\<^sub>t A \<union> to_pair ` set Di \<subseteq> setops\<^sub>s\<^sub>s\<^sub>t (Delete t s#A) \<union> to_pair ` set D"
       "setops\<^sub>s\<^sub>s\<^sub>t A \<union> to_pair ` set (filter (\<lambda>d. d \<notin> set Di) D)
          \<subseteq> setops\<^sub>s\<^sub>s\<^sub>t (Delete t s#A) \<union> to_pair ` set D"
    using sublists_set_subset[OF A''(3)] by auto
  hence **:
      "\<forall>(t,s)\<in>set (filter (\<lambda>d. d \<notin> set Di) D). \<forall>x\<in>FV t \<union> FV s. \<exists>a. \<Gamma> (Var x) = TAtom a"
      "?P A (filter (\<lambda>d. d \<notin> set Di) D)"
    using "7.prems"(3,4) by blast+
  have "eqs_tfr\<^sub>s\<^sub>t A''" using "7.IH"[OF A''(3,2) * **] by metis
  moreover have
      "\<And>u u'. u \<doteq> u' \<in> set A' \<Longrightarrow> u \<doteq> u' \<in> set A'' \<or>
          (\<exists>d \<in> set Di. u = to_pair (t,s) \<and> u' = to_pair d)"
    using A''(1) by force
  moreover have
      "\<And>X u u'. \<nexists>X\<langle>u \<doteq> u'\<rangle> \<in> set A' \<Longrightarrow> \<nexists>X\<langle>u \<doteq> u'\<rangle> \<in> set A'' \<or>
          (\<exists>d \<in> set (filter (\<lambda>d. d \<notin> set Di) D). X = [] \<and> u = to_pair (t,s) \<and> u' = to_pair d)"
    using A''(1) by force
  moreover have
      "\<forall>d\<in>set D. (\<exists>\<delta>. Unifier \<delta> (to_pair (t,s)) (to_pair d)) \<longrightarrow> \<Gamma> (to_pair (t,s)) = \<Gamma> (to_pair d)"
    using "7.prems"(4) by simp
  moreover have
      "\<forall>d \<in> set D. \<forall>x \<in> (FV (to_pair (t,s)) \<union> FV (to_pair d)). \<exists>a. \<Gamma> (Var x) = TAtom a"
    using "7.prems"(2,3) unfolding eqs_tfr\<^sub>s\<^sub>s\<^sub>t_def by auto
  ultimately show ?case using sublists_set_subset[OF A''(3)] by blast
next
  case (8 t s A D)
  then obtain d A'' where A'': "A' = to_pair (t,s) \<doteq> to_pair d#A''" "A'' \<in> tr A D" "d \<in> set D"
    by moura
  have "eqs_tfr\<^sub>s\<^sub>s\<^sub>t A" using "8.prems"(2) unfolding eqs_tfr\<^sub>s\<^sub>s\<^sub>t_def by simp
  moreover have "?P A D" using "8.prems"(4) setops\<^sub>s\<^sub>s\<^sub>t_cons_subset[of A] by fast
  ultimately have "eqs_tfr\<^sub>s\<^sub>t A''" using "8.IH"[OF A''(2) _ "8.prems"(3)] by metis
  moreover have "(\<exists>\<delta>. Unifier \<delta> (to_pair (t,s)) (to_pair d)) \<Longrightarrow> \<Gamma> (to_pair (t,s)) = \<Gamma> (to_pair d)"
    using "8.prems"(2,4) A''(3) unfolding eqs_tfr\<^sub>s\<^sub>s\<^sub>t_def by simp
  ultimately show ?case using A''(1) by fastforce
next
  case (9 X t s A D)
  let ?constr = "(map (\<lambda>d. \<nexists>X\<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>) D)"
  from 9 obtain A'' where A'': "A' = ?constr@A''" "A'' \<in> tr A D" by moura
  have "eqs_tfr\<^sub>s\<^sub>s\<^sub>t A" using "9.prems"(2) unfolding eqs_tfr\<^sub>s\<^sub>s\<^sub>t_def by simp
  moreover have "?P A D" using "9.prems"(4) setops\<^sub>s\<^sub>s\<^sub>t_cons_subset[of A] by fast
  ultimately have "eqs_tfr\<^sub>s\<^sub>t A''" using "9.IH"[OF A''(2) _ "9.prems"(3)] by metis
  moreover have
      "\<forall>d \<in> set D. \<forall>x \<in> (FV (to_pair (t,s)) \<union> FV (to_pair d))-set X. \<exists>a. \<Gamma> (Var x) = TAtom a"
    using "9.prems"(2,3) unfolding eqs_tfr\<^sub>s\<^sub>s\<^sub>t_def by auto
  ultimately show ?case using A''(1) by fastforce
qed

lemma tr_tfr:
  assumes "A' \<in> tr A []" "tfr\<^sub>s\<^sub>s\<^sub>t A"
  shows "tfr\<^sub>s\<^sub>t (trms\<^sub>s\<^sub>t A')" "eqs_tfr\<^sub>s\<^sub>t A'"
proof -
  have *: "trms\<^sub>s\<^sub>t A' \<subseteq> trms\<^sub>s\<^sub>s\<^sub>t A \<union> setops\<^sub>s\<^sub>s\<^sub>t A" using tr_trms_subset[OF assms(1)] by simp
  hence "SMP\<^sub>s\<^sub>t (trms\<^sub>s\<^sub>t A') \<subseteq> SMP\<^sub>s\<^sub>t (trms\<^sub>s\<^sub>s\<^sub>t A \<union> setops\<^sub>s\<^sub>s\<^sub>t A)" using SMP_mono by simp
  moreover have "tfr\<^sub>s\<^sub>t (trms\<^sub>s\<^sub>s\<^sub>t A \<union> setops\<^sub>s\<^sub>s\<^sub>t A)" using assms(2) unfolding tfr\<^sub>s\<^sub>s\<^sub>t_def by fast
  moreover have "Ball (trms\<^sub>s\<^sub>s\<^sub>t A \<union> setops\<^sub>s\<^sub>s\<^sub>t A) wf\<^sub>t\<^sub>r\<^sub>m" using assms(2) unfolding tfr\<^sub>s\<^sub>s\<^sub>t_def by fast
  hence "Ball (trms\<^sub>s\<^sub>t A') wf\<^sub>t\<^sub>r\<^sub>m" using * by auto
  ultimately show "tfr\<^sub>s\<^sub>t (trms\<^sub>s\<^sub>t A')" by (metis tfr_subset[OF _ *])

  have **: "eqs_tfr\<^sub>s\<^sub>s\<^sub>t A" using assms(2) unfolding tfr\<^sub>s\<^sub>s\<^sub>t_def by fast
  have "setops\<^sub>s\<^sub>s\<^sub>t A \<subseteq> SMP\<^sub>s\<^sub>t (trms\<^sub>s\<^sub>s\<^sub>t A \<union> setops\<^sub>s\<^sub>s\<^sub>t A) - Var`\<V>" using setops\<^sub>s\<^sub>s\<^sub>t_are_pairs by auto
  hence ***: "\<forall>t \<in> setops\<^sub>s\<^sub>s\<^sub>t A. \<forall>t' \<in> setops\<^sub>s\<^sub>s\<^sub>t A. (\<exists>\<delta>. Unifier \<delta> t t') \<longrightarrow> \<Gamma> t = \<Gamma> t'"
    using assms(2)  unfolding tfr\<^sub>s\<^sub>s\<^sub>t_def by blast
  show "eqs_tfr\<^sub>s\<^sub>t A'" using tr_eqs_tfr[OF assms(1) **] *** by fastforce
qed

lemma to_pair_ineqs:
  assumes "d \<cdot>p \<delta> \<cdot>p \<I> \<noteq> d' \<cdot>p \<I>"
  shows "to_pair d \<cdot> \<delta> \<cdot> \<I> \<noteq> to_pair d' \<cdot> \<I>"
proof -
  have "d \<cdot>p (\<delta> \<circ>\<^sub>s \<I>) \<noteq> d' \<cdot>p \<I>" using assms subst_pair_compose by metis
  hence "to_pair d \<cdot> (\<delta> \<circ>\<^sub>s \<I>) \<noteq> to_pair d' \<cdot> \<I>" using to_pair_eq_subst by metis
  thus ?thesis by simp
qed

lemma tr_NotInSet_constr_iff:
  fixes \<I>::"('fun,'var) subst"
  assumes "\<forall>(t,t') \<in> set D. (FV t \<union> FV t') \<inter> set X = {}"
  shows "(\<forall>\<delta>. dom\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta> = set X \<and> ground (img\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta>) \<longrightarrow> (t,s) \<cdot>p \<delta> \<cdot>p \<I> \<notin> set D \<cdot>pairs \<I>)
          \<longleftrightarrow> \<lbrakk>M; map (\<lambda>d. \<nexists>X\<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>) D\<rbrakk>\<^sub>d' \<I>"
proof -
  { assume "\<forall>\<delta>. dom\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta> = set X \<and> ground (img\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta>) \<longrightarrow> (t,s) \<cdot>p \<delta> \<cdot>p \<I> \<notin> set D \<cdot>pairs \<I>"
    with assms have "\<lbrakk>M; map (\<lambda>d. \<nexists>X\<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>) D\<rbrakk>\<^sub>d' \<I>"
    proof (induction D)
      case (Cons d D)
      obtain t' s' where d: "d = (t',s')" by moura
      have "\<lbrakk>M; map (\<lambda>d. \<nexists>X\<langle>to_pair (t, s) \<doteq> to_pair d\<rangle>) D\<rbrakk>\<^sub>d' \<I>"
           "map (\<lambda>d. \<nexists>X\<langle>to_pair (t, s) \<doteq> to_pair d\<rangle>) (d#D) =
            \<nexists>X\<langle>to_pair (t, s) \<doteq> to_pair d\<rangle>#map (\<lambda>d. \<nexists>X\<langle>to_pair (t, s) \<doteq> to_pair d\<rangle>) D"
        using Cons by auto
      moreover have
          "\<forall>\<delta>. dom\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta> = set X \<and> ground (img\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta>) \<longrightarrow> to_pair (t, s) \<cdot> \<delta> \<cdot> \<I> \<noteq> to_pair d \<cdot> \<I>"
        using to_pair_ineqs[of \<I> _ "(t,s)" d] Cons.prems(2) by auto
      moreover have "(FV t' \<union> FV s') \<inter> set X = {}" using Cons.prems(1) d by auto
      hence "\<forall>\<delta>. dom\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta> = set X \<longrightarrow> to_pair d \<cdot> \<delta> = to_pair d" using d by auto
      ultimately show ?case by simp
    qed simp
  } moreover {
    fix \<delta>::"('fun,'var) subst"
    assume "\<lbrakk>M; map (\<lambda>d. \<nexists>X\<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>) D\<rbrakk>\<^sub>d' \<I>"
      and \<delta>: "dom\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta> = set X" "ground (img\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta>)"
    with assms have "(t,s) \<cdot>p \<delta> \<cdot>p \<I> \<notin> set D \<cdot>pairs \<I>"
    proof (induction D)
      case (Cons d D)
      obtain t' s' where d: "d = (t',s')" by moura
      have "(t,s) \<cdot>p \<delta> \<cdot>p \<I> \<notin> set D \<cdot>pairs \<I>"
           "to_pair (t,s) \<cdot> \<delta> \<cdot> \<I> \<noteq> to_pair d \<cdot> \<delta> \<cdot> \<I>"
        using Cons d by auto
      moreover have "to_pair d \<cdot> \<delta> = to_pair d"
        using Cons.prems(1) to_pair_subst[of d \<delta>] d \<delta>(1) by auto
      ultimately show ?case by force
    qed simp
  } ultimately show ?thesis by metis
qed

lemma tr_Delete_constr_iff_aux1:
  assumes "\<forall>d \<in> set Di. (t,s) \<cdot>p \<I> = d \<cdot>p \<I>"
  and "\<forall>d \<in> set D - set Di. (t,s) \<cdot>p \<I> \<noteq> d \<cdot>p \<I>"
  shows "\<lbrakk>M; (map (\<lambda>d. to_pair (t,s) \<doteq> to_pair d) Di)@
             (map (\<lambda>d. \<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>) (filter (\<lambda>d. d \<notin> set Di) D))\<rbrakk>\<^sub>d' \<I>"
proof -
  from assms(2) have "\<lbrakk>M; map (\<lambda>d. \<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>) (filter (\<lambda>d. d \<notin> set Di) D)\<rbrakk>\<^sub>d' \<I>"
  proof (induction D)
    case (Cons d D)
    hence IH: "\<lbrakk>M; map (\<lambda>d. \<nexists>[]\<langle>to_pair (t, s) \<doteq> to_pair d\<rangle>) [d\<leftarrow>D . d \<notin> set Di]\<rbrakk>\<^sub>d' \<I>" by auto
    thus ?case
    proof (cases "d \<in> set Di")
      case False
      hence "(t,s) \<cdot>p \<I> \<noteq> d \<cdot>p \<I>" using Cons by simp
      hence "to_pair (t,s) \<cdot> \<I> \<noteq> to_pair d \<cdot> \<I>" using to_pair_eq_subst by metis
      moreover have "\<And>t (\<delta>::('fun,'var) subst). dom\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta> = {} \<Longrightarrow> t \<cdot> \<delta> = t" by auto
      ultimately have "\<forall>\<delta>. dom\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta> = {} \<longrightarrow> to_pair (t,s) \<cdot> \<delta> \<cdot> \<I> \<noteq> to_pair d \<cdot> \<delta> \<cdot> \<I>" by metis
      thus ?thesis using IH by simp
    qed simp
  qed simp
  moreover {
    fix B assume "\<lbrakk>M; B\<rbrakk>\<^sub>d' \<I>" 
    with assms(1) have "\<lbrakk>M; (map (\<lambda>d. to_pair (t,s) \<doteq> to_pair d) Di)@B\<rbrakk>\<^sub>d' \<I>" by (induction Di) auto
  } ultimately show ?thesis by metis
qed

lemma tr_Delete_constr_iff_aux2:
  assumes "ground M"
  and "\<lbrakk>M; (map (\<lambda>d. to_pair (t,s) \<doteq> to_pair d) Di)@
           (map (\<lambda>d. \<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>) (filter (\<lambda>d. d \<notin> set Di) D))\<rbrakk>\<^sub>d' \<I>"
  shows "(\<forall>d \<in> set Di. (t,s) \<cdot>p \<I> = d \<cdot>p \<I>) \<and> (\<forall>d \<in> set D - set Di. (t,s) \<cdot>p \<I> \<noteq> d \<cdot>p \<I>)"
proof -
  let ?c1 = "map (\<lambda>d. to_pair (t,s) \<doteq> to_pair d) Di"
  let ?c2 = "map (\<lambda>d. \<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>) (filter (\<lambda>d. d \<notin> set Di) D)"

  have "M \<cdot>set \<I> = M" using assms(1) subst_all_ground_ident by metis
  moreover have "ik\<^sub>s\<^sub>t ?c1 = {}" by auto
  ultimately have *:
       "\<lbrakk>M; map (\<lambda>d. to_pair (t,s) \<doteq> to_pair d) Di\<rbrakk>\<^sub>d' \<I>"
       "\<lbrakk>M; map (\<lambda>d. \<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>) (filter (\<lambda>d. d \<notin> set Di) D)\<rbrakk>\<^sub>d' \<I>"
    using strand_sem_split'(3,4)[of \<I> M ?c1 ?c2] assms(2) by auto
  
  from *(1) have 1: "\<forall>d \<in> set Di. (t,s) \<cdot>p \<I> = d \<cdot>p \<I>" by (induct Di) auto
  from *(2) have 2: "\<forall>d \<in> set D - set Di. (t,s) \<cdot>p \<I> \<noteq> d \<cdot>p \<I>"
  proof (induction D arbitrary: Di)
    case (Cons d D) thus ?case
    proof (cases "d \<in> set Di")
      case False
      hence IH: "\<forall>d \<in> set D - set Di. (t,s) \<cdot>p \<I> \<noteq> d \<cdot>p \<I>" using Cons by force
      have "\<And>t (\<delta>::('fun,'var) subst). dom\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta> = {} \<and> ground (img\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta>) \<longleftrightarrow> \<delta> = Var"
        unfolding ground_def img\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t_def by auto
      moreover have
          "\<forall>\<delta>. dom\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta> = {} \<and> ground (img\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta>) \<longrightarrow> to_pair (t,s) \<cdot> \<delta> \<cdot> \<I> \<noteq> to_pair d \<cdot> \<delta> \<cdot> \<I>"
        using False Cons.prems by force
      ultimately have "to_pair (t,s) \<cdot> \<I> \<noteq> to_pair d \<cdot> \<I>" by simp
      thus ?thesis using IH by force
    qed simp
  qed simp

  show ?thesis by (metis 1 2)
qed

lemma tr_Delete_constr_iff:
  fixes \<I>::"('fun,'var) subst"
  assumes "ground M"
  shows "set Di \<cdot>pairs \<I> \<subseteq> {(t,s) \<cdot>p \<I>} \<and> (t,s) \<cdot>p \<I> \<notin> (set D - set Di) \<cdot>pairs \<I> \<longleftrightarrow>
         \<lbrakk>M; (map (\<lambda>d. to_pair (t,s) \<doteq> to_pair d) Di)@
             (map (\<lambda>d. \<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>) (filter (\<lambda>d. d \<notin> set Di) D))\<rbrakk>\<^sub>d' \<I>"
proof -
  let ?constr = "(map (\<lambda>d. to_pair (t,s) \<doteq> to_pair d) Di)@
                 (map (\<lambda>d. \<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>) (filter (\<lambda>d. d \<notin> set Di) D))"
  { assume "set Di \<cdot>pairs \<I> \<subseteq> {(t,s) \<cdot>p \<I>}" "(t,s) \<cdot>p \<I> \<notin> (set D - set Di) \<cdot>pairs \<I>"
    hence "\<forall>d \<in> set Di. (t,s) \<cdot>p \<I> = d \<cdot>p \<I>" "\<forall>d \<in> set D - set Di. (t,s) \<cdot>p \<I> \<noteq> d \<cdot>p \<I>"
      by auto
    hence "\<lbrakk>M; ?constr\<rbrakk>\<^sub>d' \<I>" using tr_Delete_constr_iff_aux1 by simp
  } moreover {
    assume "\<lbrakk>M; ?constr\<rbrakk>\<^sub>d' \<I>"
    hence "\<forall>d \<in> set Di. (t,s) \<cdot>p \<I> = d \<cdot>p \<I>" "\<forall>d \<in> set D - set Di. (t,s) \<cdot>p \<I> \<noteq> d \<cdot>p \<I>"
      using assms tr_Delete_constr_iff_aux2 by auto
    hence "set Di \<cdot>pairs \<I> \<subseteq> {(t,s) \<cdot>p \<I>} \<and> (t,s) \<cdot>p \<I> \<notin> (set D - set Di) \<cdot>pairs \<I>" by force
  } ultimately show ?thesis by metis
qed

lemma tr_sem_equiv:
  assumes "\<forall>(t,t') \<in> set D. (FV t \<union> FV t') \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}"
  and "FV\<^sub>s\<^sub>s\<^sub>t A \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}" "ground M"
  and \<I>: "interpretation\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>"
  shows "\<lbrakk>M; set D \<cdot>pairs \<I>; A\<rbrakk>\<^sub>s \<I> \<longleftrightarrow> (\<exists>A' \<in> tr A D. \<lbrakk>M; A'\<rbrakk>\<^sub>d \<I>)" (is "?P \<longleftrightarrow> ?Q")
proof
  have M\<I>: "M \<cdot>set \<I> = M" by (rule subst_all_ground_ident[OF assms(3)])
  have \<I>_grounds: "\<And>t. FV (t \<cdot> \<I>) = {}" by (rule interpretation_grounds[OF \<I>])
  have "?P \<Longrightarrow> (\<exists>A' \<in> tr A D. \<lbrakk>M; A'\<rbrakk>\<^sub>d' \<I>)" using assms(1,2,3)
  proof (induction A arbitrary: D rule: strand_sem_stateful_induct)
    case (Nil M D) thus ?case by simp
  next
    case (ConsRcv M D t A)
    hence "\<lbrakk>insert (t \<cdot> \<I>) M; set D \<cdot>pairs \<I>; A\<rbrakk>\<^sub>s \<I>"
          "\<forall>(t,t') \<in> set D. (FV t \<union> FV t') \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}"
          "FV\<^sub>s\<^sub>s\<^sub>t A \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}" "ground (insert (t \<cdot> \<I>) M)"
      using \<I> unfolding FV\<^sub>s\<^sub>s\<^sub>t_def bvars\<^sub>s\<^sub>s\<^sub>t_def by auto
    then obtain A' where A': "A' \<in> tr A D" "\<lbrakk>insert (t \<cdot> \<I>) M; A'\<rbrakk>\<^sub>d' \<I>" by (metis ConsRcv.IH)
    thus ?case by auto
  next
    case (ConsSnd M D t A)
    hence "\<lbrakk>M; set D \<cdot>pairs \<I>; A\<rbrakk>\<^sub>s \<I>"
          "\<forall>(t,t') \<in> set D. (FV t \<union> FV t') \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}"
          "FV\<^sub>s\<^sub>s\<^sub>t A \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}" "ground M"
      and *: "M \<turnstile> t \<cdot> \<I>"
      using \<I> unfolding FV\<^sub>s\<^sub>s\<^sub>t_def bvars\<^sub>s\<^sub>s\<^sub>t_def by auto
    then obtain A' where A': "A' \<in> tr A D" "\<lbrakk>M; A'\<rbrakk>\<^sub>d' \<I>" by (metis ConsSnd.IH)
    thus ?case using * by auto
  next
    case (ConsEq M D t t' A)
    hence "\<lbrakk>M; set D \<cdot>pairs \<I>; A\<rbrakk>\<^sub>s \<I>"
          "\<forall>(t,t') \<in> set D. (FV t \<union> FV t') \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}"
          "FV\<^sub>s\<^sub>s\<^sub>t A \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}" "ground M"
      and *: "t \<cdot> \<I> = t' \<cdot> \<I>"
      using \<I> unfolding FV\<^sub>s\<^sub>s\<^sub>t_def bvars\<^sub>s\<^sub>s\<^sub>t_def by auto
    then obtain A' where A': "A' \<in> tr A D" "\<lbrakk>M; A'\<rbrakk>\<^sub>d' \<I>" by (metis ConsEq.IH)
    thus ?case using * by auto
  next
    case (ConsIneq M D X t t' A)
    hence "\<lbrakk>M; set D \<cdot>pairs \<I>; A\<rbrakk>\<^sub>s \<I>"
          "\<forall>(t,t') \<in> set D. (FV t \<union> FV t') \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}"
          "FV\<^sub>s\<^sub>s\<^sub>t A \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}" "ground M"
      and *: "\<forall>\<delta>. dom\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta> = set X \<and> ground (img\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta>) \<longrightarrow> t \<cdot> \<delta> \<cdot> \<I> \<noteq> t' \<cdot> \<delta> \<cdot> \<I>"
      using \<I> unfolding FV\<^sub>s\<^sub>s\<^sub>t_def bvars\<^sub>s\<^sub>s\<^sub>t_def by force+
    then obtain A' where A': "A' \<in> tr A D" "\<lbrakk>M; A'\<rbrakk>\<^sub>d' \<I>" by (metis ConsIneq.IH)
    thus ?case using * by auto
  next
    case (ConsIns M D t s A)
    hence "\<lbrakk>M; set (List.insert (t,s) D) \<cdot>pairs \<I>; A\<rbrakk>\<^sub>s \<I>"
          "\<forall>(t,t') \<in> set (List.insert (t,s) D). (FV t \<union> FV t') \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}"
          "FV\<^sub>s\<^sub>s\<^sub>t A \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}" "ground M"
      unfolding FV\<^sub>s\<^sub>s\<^sub>t_def bvars\<^sub>s\<^sub>s\<^sub>t_def by auto
    then obtain A' where A': "A' \<in> tr A (List.insert (t,s) D)" "\<lbrakk>M; A'\<rbrakk>\<^sub>d' \<I>" by (metis ConsIns.IH)
    thus ?case by auto
  next
    case (ConsDel M D t s A)
    hence *: "\<lbrakk>M; (set D \<cdot>pairs \<I>) - {(t,s) \<cdot>p \<I>}; A\<rbrakk>\<^sub>s \<I>"
             "\<forall>(t,t')\<in>set D. (FV t \<union> FV t') \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}"
             "FV\<^sub>s\<^sub>s\<^sub>t A \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}" "ground M"
      unfolding FV\<^sub>s\<^sub>s\<^sub>t_def bvars\<^sub>s\<^sub>s\<^sub>t_def by auto
    then obtain Di where Di:
        "Di \<subseteq> set D" "Di \<cdot>pairs \<I> \<subseteq> {(t,s) \<cdot>p \<I>}" "(t,s) \<cdot>p \<I> \<notin> (set D - Di) \<cdot>pairs \<I>"
      using subset_subst_pairs_diff_exists'[of "set D"] by moura
    hence **: "(set D \<cdot>pairs \<I>) - {(t,s) \<cdot>p \<I>} = (set D - Di) \<cdot>pairs \<I>" by blast

    obtain Di' where Di': "set Di' = Di" "Di' \<in> set (sublists D)"
      using subset_sublist_exists[OF Di(1)] by moura
    
    let ?constr = "map (\<lambda>d. to_pair (t,s) \<doteq> to_pair d) Di'@
                   map (\<lambda>d. \<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>) (filter (\<lambda>d. d \<notin> set Di') D)"

    from Di' have ***:
        "(set D \<cdot>pairs \<I>) - {(t,s) \<cdot>p \<I>} = (set (filter (\<lambda>d. d \<notin> set Di') D) \<cdot>pairs \<I>)"
      using Di ** by auto
    have ****: "\<forall>(t,t')\<in>set (filter (\<lambda>d. d \<notin> set Di') D). (FV t \<union> FV t') \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}"
      using *(2) Di(1) Di'(1) sublists_set_subset[OF Di'(2)] by simp
    have "set D - Di = set (filter (\<lambda>d. d \<notin> set Di') D)" using Di Di' by auto
    hence *****: "\<lbrakk>M; set (filter (\<lambda>d. d \<notin> set Di') D) \<cdot>pairs \<I>; A\<rbrakk>\<^sub>s \<I>"
      using *(1) ** by metis
    obtain A' where A': "A' \<in> tr A (filter (\<lambda>d. d \<notin> set Di') D)" "\<lbrakk>M; A'\<rbrakk>\<^sub>d' \<I>"
      using ConsDel.IH[OF ***** **** *(3,4)] by moura
    hence constr_sat: "\<lbrakk>M; ?constr\<rbrakk>\<^sub>d' \<I>"
      using Di Di' *(1) *** tr_Delete_constr_iff[OF *(4), of \<I> Di' t s D] 
      by auto

    have "?constr@A' \<in> tr (Delete t s#A) D" using A'(1) Di' by auto
    moreover have "ik\<^sub>s\<^sub>t ?constr = {}" by auto
    hence "\<lbrakk>M \<cdot>set \<I>; ?constr\<rbrakk>\<^sub>d' \<I>" "\<lbrakk>M \<union> ik\<^sub>s\<^sub>t ?constr \<cdot>set \<I>; A'\<rbrakk>\<^sub>d' \<I>"
      using constr_sat A'(2) subst_all_ground_ident[OF *(4)] by simp_all
    ultimately show ?case by (metis strand_sem_append'(2) subst_all_ground_ident[OF *(4)])
  next
    case (ConsIn M D t s A)
    hence "\<lbrakk>M; set D \<cdot>pairs \<I>; A\<rbrakk>\<^sub>s \<I>"
          "\<forall>(t,t') \<in> set D. (FV t \<union> FV t') \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}"
          "FV\<^sub>s\<^sub>s\<^sub>t A \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}" "ground M"
      and *: "(t,s) \<cdot>p \<I> \<in> set D \<cdot>pairs \<I>"
      using \<I> unfolding FV\<^sub>s\<^sub>s\<^sub>t_def bvars\<^sub>s\<^sub>s\<^sub>t_def by auto
    then obtain A' where A': "A' \<in> tr A D" "\<lbrakk>M; A'\<rbrakk>\<^sub>d' \<I>" by (metis ConsIn.IH)
    moreover obtain d where "d \<in> set D" "to_pair (t,s) \<cdot> \<I> = to_pair d \<cdot> \<I>" using * by auto
    ultimately show ?case using * by auto
  next
    case (ConsNotin M D X t s A)
    let ?ineqs = "map (\<lambda>d. \<nexists>X\<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>) D"
    have 1: "\<lbrakk>M; set D \<cdot>pairs \<I>; A\<rbrakk>\<^sub>s \<I>" "ground M" using ConsNotin by auto
    have 2: "\<forall>(t,t') \<in> set D. (FV t \<union> FV t') \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}" "FV\<^sub>s\<^sub>s\<^sub>t A \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}" 
      using ConsNotin \<I> unfolding FV\<^sub>s\<^sub>s\<^sub>t_def bvars\<^sub>s\<^sub>s\<^sub>t_def by force+
    have *: "\<forall>\<delta>. dom\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta> = set X \<and> ground (img\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta>) \<longrightarrow> (t,s) \<cdot>p \<delta> \<cdot>p \<I> \<notin> set D \<cdot>pairs \<I>"
      using ConsNotin.prems(1) by simp
    from 1 2 obtain A' where A': "A' \<in> tr A D" "\<lbrakk>M; A'\<rbrakk>\<^sub>d' \<I>" by (metis ConsNotin.IH)
    moreover have "\<forall>(t,t') \<in> set D. (FV t \<union> FV t') \<inter> set X = {}"
      using ConsNotin.prems(2) unfolding bvars\<^sub>s\<^sub>s\<^sub>t_def by auto
    hence "\<lbrakk>M; ?ineqs\<rbrakk>\<^sub>d' \<I>" using * tr_NotInSet_constr_iff by metis
    moreover have "ik\<^sub>s\<^sub>t ?ineqs = {}" by auto
    moreover have "M \<cdot>set \<I> = M" using 1(2) \<I> by (simp add: subst_all_ground_ident)
    ultimately show ?case
      using strand_sem_append'(2)[of \<I> M ?ineqs A']
      by force
  qed
  thus "?P \<Longrightarrow> ?Q" by (metis M\<I> strand_sem_eq_defs(2))

  have "(\<exists>A' \<in> tr A D. \<lbrakk>M; A'\<rbrakk>\<^sub>d' \<I>) \<Longrightarrow> ?P" using assms(1,2,3)
  proof (induction A arbitrary: D rule: strand_sem_stateful_induct)
    case (Nil M D) thus ?case by simp
  next
    case (ConsRcv M D t A)
    hence "\<exists>A' \<in> tr A D. \<lbrakk>insert (t \<cdot> \<I>) M; A'\<rbrakk>\<^sub>d' \<I>"
          "\<forall>(t,t') \<in> set D. (FV t \<union> FV t') \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}"
          "FV\<^sub>s\<^sub>s\<^sub>t A \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}" "ground (insert (t \<cdot> \<I>) M)"
      using \<I> unfolding FV\<^sub>s\<^sub>s\<^sub>t_def bvars\<^sub>s\<^sub>s\<^sub>t_def by auto
    hence "\<lbrakk>insert (t \<cdot> \<I>) M; set D \<cdot>pairs \<I>; A\<rbrakk>\<^sub>s \<I>" by (metis ConsRcv.IH)
    thus ?case by auto
  next
    case (ConsSnd M D t A)
    hence "\<exists>A' \<in> tr A D. \<lbrakk>M; A'\<rbrakk>\<^sub>d' \<I>"
          "\<forall>(t,t') \<in> set D. (FV t \<union> FV t') \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}"
          "FV\<^sub>s\<^sub>s\<^sub>t A \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}" "ground M"
      and *: "M \<turnstile> t \<cdot> \<I>"
      using \<I> unfolding FV\<^sub>s\<^sub>s\<^sub>t_def bvars\<^sub>s\<^sub>s\<^sub>t_def by auto
    hence "\<lbrakk>M; set D \<cdot>pairs \<I>; A\<rbrakk>\<^sub>s \<I>" by (metis ConsSnd.IH)
    thus ?case using * by auto
  next
    case (ConsEq M D t t' A)
    hence "\<exists>A' \<in> tr A D. \<lbrakk>M; A'\<rbrakk>\<^sub>d' \<I>"
          "\<forall>(t,t') \<in> set D. (FV t \<union> FV t') \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}"
          "FV\<^sub>s\<^sub>s\<^sub>t A \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}" "ground M"
      and *: "t \<cdot> \<I> = t' \<cdot> \<I>"
      using \<I> unfolding FV\<^sub>s\<^sub>s\<^sub>t_def bvars\<^sub>s\<^sub>s\<^sub>t_def by auto
    hence "\<lbrakk>M; set D \<cdot>pairs \<I>; A\<rbrakk>\<^sub>s \<I>" by (metis ConsEq.IH)
    thus ?case using * by auto
  next
    case (ConsIneq M D X t t' A)
    hence "\<exists>A' \<in> tr A D. \<lbrakk>M; A'\<rbrakk>\<^sub>d' \<I>"
          "\<forall>(t,t') \<in> set D. (FV t \<union> FV t') \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}"
          "FV\<^sub>s\<^sub>s\<^sub>t A \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}" "ground M"
      and *: "\<forall>\<delta>. dom\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta> = set X \<and> ground (img\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta>) \<longrightarrow> t \<cdot> \<delta> \<cdot> \<I> \<noteq> t' \<cdot> \<delta> \<cdot> \<I>"
      using \<I> unfolding FV\<^sub>s\<^sub>s\<^sub>t_def bvars\<^sub>s\<^sub>s\<^sub>t_def by force+
    hence "\<lbrakk>M; set D \<cdot>pairs \<I>; A\<rbrakk>\<^sub>s \<I>" by (metis ConsIneq.IH)
    thus ?case using * by auto
  next
    case (ConsIns M D t s A)
    hence "\<exists>A' \<in> tr A (List.insert (t,s) D). \<lbrakk>M; A'\<rbrakk>\<^sub>d' \<I>"
          "\<forall>(t,t') \<in> set (List.insert (t,s) D). (FV t \<union> FV t') \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}"
          "FV\<^sub>s\<^sub>s\<^sub>t A \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}" "ground M"
      unfolding FV\<^sub>s\<^sub>s\<^sub>t_def bvars\<^sub>s\<^sub>s\<^sub>t_def by auto
    hence "\<lbrakk>M; set (List.insert (t,s) D) \<cdot>pairs \<I>; A\<rbrakk>\<^sub>s \<I>" by (metis ConsIns.IH)
    thus ?case by auto
  next
    case (ConsDel M D t s A)
    let ?constr = "\<lambda>Di. map (\<lambda>d. to_pair (t,s) \<doteq> to_pair d) Di@
                        map (\<lambda>d. \<nexists>[]\<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>) (filter (\<lambda>d. d \<notin> set Di) D)"
    let ?flt = "\<lambda>Di. filter (\<lambda>d. d \<notin> set Di) D"

    obtain A' Di where A':
        "?constr Di@A' \<in> tr (Delete t s#A) D"
        "A' \<in> tr A (?flt Di)"
        "Di \<in> set (sublists D)"
        "\<lbrakk>M; ?constr Di@A'\<rbrakk>\<^sub>d' \<I>"
      using ConsDel.prems(1) by moura

    have 1: "\<forall>(t,t')\<in>set (?flt Di). (FV t \<union> FV t') \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}" using ConsDel.prems(2) by auto
    have 2: "FV\<^sub>s\<^sub>s\<^sub>t A \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}" using ConsDel.prems(3) by auto
    have "ik\<^sub>s\<^sub>t (?constr Di) = {}" by auto
    hence 3: "\<lbrakk>M; A'\<rbrakk>\<^sub>d' \<I>"
      using subst_all_ground_ident[OF ConsDel.prems(4)] A'(4)
            strand_sem_split'(4)[of \<I> M "?constr Di" A']
      by simp
    have IH: "\<lbrakk>M; set (?flt Di) \<cdot>pairs \<I>; A\<rbrakk>\<^sub>s \<I>"
      by (metis ConsDel.IH[OF _ 1 2 ConsDel.prems(4)] 3 A'(2))

    have "\<lbrakk>M; ?constr Di\<rbrakk>\<^sub>d' \<I>"
      using subst_all_ground_ident[OF ConsDel.prems(4)] strand_sem_split'(3) A'(4)
      by metis
    hence *: "set Di \<cdot>pairs \<I> \<subseteq> {(t,s) \<cdot>p \<I>}" "(t,s) \<cdot>p \<I> \<notin> (set D - set Di) \<cdot>pairs \<I>"
      using tr_Delete_constr_iff[OF ConsDel.prems(4), of \<I> Di t s D] by auto
    have 4: "set (?flt Di) \<cdot>pairs \<I> = (set D \<cdot>pairs \<I>) - {((t,s) \<cdot>p \<I>)}"
    proof
      show "set (?flt Di) \<cdot>pairs \<I> \<subseteq> (set D \<cdot>pairs \<I>) - {((t,s) \<cdot>p \<I>)}"
      proof
        fix u u' assume u: "(u,u') \<in> set (?flt Di) \<cdot>pairs \<I>"
        then obtain v v' where v: "(v,v') \<in> set D - set Di" "(v,v') \<cdot>p \<I> = (u,u')" by auto
        hence "(u,u') \<noteq> (t,s) \<cdot>p \<I>" using * by force
        thus "(u,u') \<in>  (set D \<cdot>pairs \<I>) - {((t,s) \<cdot>p \<I>)}"
          using u v * sublists_set_subset[OF A'(3)] by auto
      qed
      show "(set D \<cdot>pairs \<I>) - {((t,s) \<cdot>p \<I>)} \<subseteq> set (?flt Di) \<cdot>pairs \<I>"
        using * sublists_set_subset[OF A'(3)] by force
    qed

    show ?case using 4 IH by simp
  next
    case (ConsIn M D t s A)
    hence "\<exists>A' \<in> tr A D. \<lbrakk>M; A'\<rbrakk>\<^sub>d' \<I>"
          "\<forall>(t,t') \<in> set D. (FV t \<union> FV t') \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}"
          "FV\<^sub>s\<^sub>s\<^sub>t A \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}" "ground M"
      and *: "(t,s) \<cdot>p \<I> \<in> set D \<cdot>pairs \<I>"
      using \<I> unfolding FV\<^sub>s\<^sub>s\<^sub>t_def bvars\<^sub>s\<^sub>s\<^sub>t_def by force+
    hence "\<lbrakk>M; set D \<cdot>pairs \<I>; A\<rbrakk>\<^sub>s \<I>" by (metis ConsIn.IH)
    moreover obtain d where "d \<in> set D" "to_pair (t,s) \<cdot> \<I> = to_pair d \<cdot> \<I>" using * by auto
    ultimately show ?case using * by auto
  next
    case (ConsNotin M D X t s A)
    let ?ineqs = "map (\<lambda>d. \<nexists>X\<langle>to_pair (t,s) \<doteq> to_pair d\<rangle>) D"
    obtain B where B: "?ineqs@B \<in> tr (NotInSet X t s#A) D" "\<lbrakk>M; ?ineqs@B\<rbrakk>\<^sub>d' \<I>" "B \<in> tr A D"
      using ConsNotin.prems(1) by moura
    moreover have "M \<cdot>set \<I> = M" using ConsNotin.prems(4) \<I> by (simp add: subst_all_ground_ident)
    moreover have "ik\<^sub>s\<^sub>t ?ineqs = {}" by auto
    ultimately have "\<lbrakk>M; B\<rbrakk>\<^sub>d' \<I>" using strand_sem_split'(4)[of \<I> M ?ineqs B] by simp
    moreover have "\<forall>(t,t')\<in>set D. (FV t \<union> FV t') \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}" "FV\<^sub>s\<^sub>s\<^sub>t A \<inter> bvars\<^sub>s\<^sub>s\<^sub>t A = {}"
      using ConsNotin.prems(2,3) unfolding FV\<^sub>s\<^sub>s\<^sub>t_def bvars\<^sub>s\<^sub>s\<^sub>t_def by force+
    ultimately have "\<lbrakk>M; set D \<cdot>pairs \<I>; A\<rbrakk>\<^sub>s \<I>" by (metis ConsNotin.IH B(3) ConsNotin.prems(4))
    moreover have "\<forall>(t, t')\<in>set D. (FV t \<union> FV t') \<inter> set X = {}"
      using ConsNotin.prems(2) unfolding bvars\<^sub>s\<^sub>s\<^sub>t_def by force
    ultimately show ?case
      using tr_NotInSet_constr_iff[of D X \<I> t s M] B(2)
            strand_sem_split'(3)[of \<I> M ?ineqs B] \<open>M \<cdot>set \<I> = M\<close>
      by simp
  qed
  thus "?Q \<Longrightarrow> ?P" by (metis M\<I> strand_sem_eq_defs(2))
qed

theorem wt_attack_if_tfr_attack_stateful:
  assumes "wf\<^sub>s\<^sub>s\<^sub>t {} \<A>" "FV\<^sub>s\<^sub>s\<^sub>t \<A> \<inter> bvars\<^sub>s\<^sub>s\<^sub>t \<A> = {}" "tfr\<^sub>s\<^sub>s\<^sub>t \<A>"
  and "interpretation\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>" "\<lbrakk>{}; {}; \<A>\<rbrakk>\<^sub>s \<I>"
  shows "\<exists>\<I>\<^sub>\<tau>. interpretation\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>\<^sub>\<tau> \<and> \<lbrakk>{}; {}; \<A>\<rbrakk>\<^sub>s \<I>\<^sub>\<tau> \<and> wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>\<^sub>\<tau> \<and> (\<forall>x. wf\<^sub>t\<^sub>r\<^sub>m (\<I>\<^sub>\<tau> x))"
proof -
  obtain \<A>' where \<A>': "\<A>' \<in> tr \<A> []" "\<lbrakk>{}; \<A>'\<rbrakk>\<^sub>d \<I>"
    using tr_sem_equiv[of "[]" \<A> "{}"] assms(2,4,5) by auto
  have *: "wf\<^sub>s\<^sub>t {} \<A>'" "FV\<^sub>s\<^sub>t \<A>' \<inter> bvars\<^sub>s\<^sub>t \<A>' = {}" "tfr\<^sub>s\<^sub>t (trms\<^sub>s\<^sub>t \<A>')" "eqs_tfr\<^sub>s\<^sub>t \<A>'"
    using tr_wf[OF _ _ assms(1,2) \<A>'(1)]
          tr_tfr[OF \<A>'(1) assms(3)]
          tr_vars_disj[OF \<A>'(1) _ assms(2)]
    by force+
  obtain \<I>\<^sub>\<tau> where \<I>\<^sub>\<tau>: "interpretation\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>\<^sub>\<tau>" "\<lbrakk>{}; \<A>'\<rbrakk>\<^sub>d \<I>\<^sub>\<tau>" "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>\<^sub>\<tau>" "\<forall>x. wf\<^sub>t\<^sub>r\<^sub>m (\<I>\<^sub>\<tau> x)"
    using wt_attack_if_tfr_attack_d[OF * Ana_invar_subst' assms(4)] \<A>'(2) by moura
  thus ?thesis using tr_sem_equiv[of "[]" \<A> "{}"] assms(2,4) \<A>'(1) by auto
qed

end

end
